import logging as loggingmod
import os

# from biothings import config
from biothings.utils.loggers import get_logger


def upload_worker(name, storage_class, loaddata_func, col_name, batch_size, batch_num, *args, **kwargs):
    """
    Pickable job launcher, typically running from multiprocessing.
    storage_class will instanciate with col_name, the destination
    collection name. loaddata_func is the parsing/loading function,
    called with `*args`.
    """
    data = []
    db = kwargs.get("db", None)
    max_batch_num = kwargs.get("max_batch_num", None)
    try:
        data = loaddata_func(*args)
        if type(storage_class) is tuple:
            klass_name = "_".join([k.__class__.__name__ for k in storage_class])
            storage = type(klass_name, storage_class, {})(None, col_name, loggingmod)
        else:
            storage = storage_class(db, col_name, loggingmod)
        return storage.process(data, batch_size, max_batch_num)
    except Exception as e:
        logger_name = "%s_batch_%s" % (name, batch_num)
        logger, logfile = get_logger(logger_name)
        logger.exception(e)
        logger.error(
            "Parameters:\nname=%s\nstorage_class=%s\n" % (name, storage_class)
            + "loaddata_func=%s\ncol_name=%s\nbatch_size=%s\n" % (loaddata_func, col_name, batch_size)
            + "args=%s" % repr(args)
        )
        import pickle

        # when logfile is None, we're in CLI mode, so we just use the current directory
        pickfile = os.path.join(os.path.dirname(logfile) if logfile else ".", "%s.pick" % logger_name)
        try:
            pickle.dump(
                {
                    "exc": e,
                    "params": {"name": name, "storage_class": storage_class},
                    "loaddata_func": repr(loaddata_func),  # loaddata_func may not be pickle-able
                    "col_name": col_name,
                    "batch_size": batch_size,
                    "args": args,
                },
                open(pickfile, "wb"),
            )
        except (TypeError, pickle.PicklingError) as ie:
            logger.warning("Could not pickle batch errors: %s", ie)
        raise e



import logging
from datetime import datetime, timezone
from time import time

from biothings.utils.common import merge, timesofar

# NO CONCURRENT
# TASK SUPPORT YET

_map = {
    # "pre": PreSnapshotState,
    # "snapshot": MainSnapshotState,
    # "post": PostSnapshotState
}


def dispatch(step):
    return _map[step]


def audit(src_build, logger=None):
    if not logger:
        logger = logging.getLogger(__name__)

    for build in src_build.find():
        for num, job in enumerate(build.get("jobs", [])):
            if job.get("status") == "in progress":
                job["status"] = "canceled"

                msg = "<Job #%d in <Build '%s'> cancelled>"
                logger.warning(msg, num, build["_id"])

        src_build.replace_one({"_id": build["_id"]}, build)


class _TaskState:
    name = NotImplemented  # string representation of the step
    step = NotImplemented  # job registration display notation
    func = NotImplemented  # method name of the step to run
    regx = False  # ready to update the build doc

    def __init__(self, col, _id):
        self._col = col
        self._id = _id

    @classmethod
    def __init_subclass__(cls):
        _map[cls.name] = cls

    def started(self, **extras):
        timestamp = datetime.now().astimezone()
        self._col.update(
            {"_id": self._id},
            {
                "$push": {
                    "jobs": {
                        "step": self.step,
                        "status": "in progress",
                        "step_started_at": timestamp,
                        **extras,
                    }
                }
            },
        )

    def _finished(self, _doc, _job):
        doc = self._col.find_one({"_id": self._id})
        job = doc["jobs"][-1]
        t0 = job["step_started_at"]

        # New pymongo lib version changed the default configuration value of `tz_aware` to `False` instead of `True`.
        # Because of that the datetime value read from mongodb was not including the timezone anymore.
        # The solution is forcing the datetime value to include the UTC timezone when it is missing.
        # Reference: https://pymongo.readthedocs.io/en/stable/migrate-to-pymongo4.html?highlight=datetime#tz-aware-defaults-to-false
        if not t0.tzinfo:
            t0 = t0.replace(tzinfo=timezone.utc)
        t0 = t0.timestamp()

        job["time_in_s"] = round(time() - t0, 0)
        job["time"] = timesofar(t0)

        if self.regx:
            merge(doc, _doc)
        merge(job, _job)

        self._col.replace_one({"_id": self._id}, doc)

    def failed(self, dBuild, **dJob):
        self._finished({"snapshot": dBuild}, {"status": "failed", **dJob})

    def succeed(self, dBuild, **dJob):
        self._finished({"snapshot": dBuild}, {"status": "success", **dJob})

    def __str__(self):
        return f"<{type(self).__name__} {self._id}>"


class PreSnapshotState(_TaskState):
    name = "pre"
    step = "pre-snapshot"
    func = "pre_snapshot"


class MainSnapshotState(_TaskState):
    name = "snapshot"
    step = "snapshot"
    func = "_snapshot"
    regx = True


class PostSnapshotState(_TaskState):
    name = "post"
    step = "post-snapshot"
    func = "post_snapshot"
    regx = True


def test():
    from pymongo import MongoClient

    client = MongoClient()
    collection = client["biothings"]["src_build"]
    state = PreSnapshotState(collection, "mynews_202105261855_5ffxvchx")
    state.started(logfile="__TEST_LOCATION__")
    state.succeed({"test_snapshot_01": {}})

    state.started(logfile="__TEST_LOCATION__")
    state.failed(ValueError("__TEST_EXCEPTION__"))


if __name__ == "__main__":
    test()



class AutoHubValidateError(Exception):
    reason = None

    def __init__(self, reason, *args, **kwargs):
        self.reason = reason
        super().__init__(*args, **kwargs)


class AutoHubValidator:
    """
    This class aims to provide an easy way to customize validation logic for installing a hub from a release.
    """

    def __init__(self, auto_hub_feature):
        self.auto_hub_feature = auto_hub_feature

    def validate(self, force=False, **kwargs):
        """
        Check if the release is valid to install.
        If invalid, it should raise an AutoHubValidateError include any reason to stop the progress.
        """
        pass



import datetime
import itertools
import logging
from collections import UserDict, UserList
from dataclasses import dataclass
from pprint import pformat

from elasticsearch import AsyncElasticsearch
from pymongo.collection import Collection

# NOTE
# Throughout this module, an XML-like serialization
# format is utilized, just like the python default:
#
# >>> object()
# <object object at 0x7f7b4882a110>
#
# This format helps preserve hierarchical structure
# in the object it represents, is intuitive to read,
# and closely represent the underlying programming
# concepts, each tagname corresponds to one class.
#
_TAB = " " * 2


class _Index(UserDict):
    def __str__(self):
        return (
            f"<Index"
            f" env={repr(self.get('environment'))}"
            f" name={repr(self.get('_id'))}"
            f" ts={repr(self.get('created_at'))}"
            f"/>"
        )


class _Indices(UserList):
    group = NotImplemented

    def __init__(self, initlist=None):
        super().__init__(initlist)
        self.data = [_Index(x) for x in self]

    def __str__(self):
        return "\n".join(
            (f"<{self.group} len={len(self)}>", *(_TAB + str(index) for index in self), f"<{self.group}/>")
        )


class _IndicesToKeep(_Indices):
    group = "Keep"


class _IndicesToRemove(_Indices):
    group = "Remove"


@dataclass
class _BuildConfig:
    name: str
    remove: _IndicesToRemove
    keep: _IndicesToKeep

    def __str__(self):
        return "\n".join(
            (
                f"<BuildConfig {repr(self.name)}>",
                *(_TAB + line for line in str(self.remove).split("\n")),
                *(_TAB + line for line in str(self.keep).split("\n")),
                "<BuildConfig/>",
            )
        )

    def __iter__(self):
        return iter(self.remove)


class _CleanUps(UserList):
    def __str__(self):
        lines = map(lambda x: str(x).split("\n"), self)
        lines = itertools.chain.from_iterable(lines)
        return "\n".join(("<CleanUps>", *(_TAB + line for line in lines), "<CleanUps/>"))


# OUTPUT EXAMPLE
# ---------------------
# >>> _CleanUpList(...)
# <CleanUp>
#   <BuildConfig 'mygene_allspecies'>
#     <Remove len=6>
#       <Index env='prod' name='mygene_xbuo6d' ts=datetime(2019, 8, 27, ...)/>
#       <Index env='prod' name='mygene_vxia0r' ts=datetime(2020, 1, 8, ...)/>
#       <Index env='prod' name='mygene_25wlt4' ts=datetime(2020, 1, 20, ...)/>
#       <Index env='local' name='mygene_ufkw79' ts=datetime(2020, 8, 28, ...)/>
#       <Index env='local' name='mygene_uq3chc' ts=datetime(2021, 4, 6, ...)/>
#       <Index env='local' name='mygene_mnkct5' ts=datetime(2021, 4, 15, ...)/>
#     <Remove/>
#     <Keep len=3>
#       <Index env='su10' name='mygene_ibjpha' ts=datetime(2021, 6, 30, ...)/>
#       <Index env='su10' name='mygene_osyzmt' ts=datetime(2021, 8, 9, ...)/>
#       <Index env='su10' name='mygene_test' ts=datetime(2021, 8, 9, ...)/>
#     <Keep/>
#   <BuildConfig/>
#   <BuildConfig 'demo_allspecies'>
#     <Remove len=3>
#       <Index env='test' name='demo_mygene_ngupjv' ts=datetime(2018, 3, 12, ...)/>
#       <Index env='test' name='demo_mygene_rpguqe' ts=datetime(2018, 8, 6, ...)/>
#       <Index env='test' name='demo_mygene_vqpur6' ts=datetime(2019, 1, 29, ...)/>
#     <Remove/>
#     <Keep len=3>
#       <Index env='test' name='demo_mygene_irvwa0' ts=datetime(2019, 3, 25, ...)/>
#       <Index env='test' name='demo_mygene_cpoldl' ts=datetime(2020, 1, 8, ...)/>
#       <Index env='local' name='demo_mygene_qekeic' ts=datetime(2020, 4, 22, ...)/>
#     <Keep/>
#   <BuildConfig/>
# <CleanUp/>
#


class CleanUpResult(list):
    def __repr__(self):
        return "".join((type(self).__name__, "(", "\n" if self else "", pformat(list(self), width=150), ")"))


class Cleaner:
    def __init__(self, collection, indexers, logger=None):
        self.collection = collection  # pymongo.collection.Collection
        self.indexers = indexers  # hub.dataindex.IndexManager
        self.logger = logger or logging.getLogger(__name__)

    def find(self, env=None, keep=3, **filters):
        if not isinstance(self.collection, Collection):
            raise NotImplementedError("Require MongoDB Hubdb.")

        results = list(
            self.collection.aggregate(
                [
                    {"$project": {"build_config": "$build_config._id", "index": {"$objectToArray": "$index"}}},
                    {"$unwind": {"path": "$index"}},
                    {"$addFields": {"index.v.build_config": "$build_config", "index.v._id": "$index.k"}},
                    {"$replaceRoot": {"newRoot": "$index.v"}},
                    {"$match": {"environment": env or {"$exists": True}, **filters}},
                    # {X
                    # .  '_id': 'mynews_202012280220_vsdevjdk',
                    # ...'build_config': 'mynews',  ──────────┐
                    # ...'environment': 'local',              │
                    # :  'created_at': datetime(...)          │
                    # }Y                                      │
                    # {                              GROUP BY │
                    #    '_id': 'mynews',  <──────────────────┘
                    # ...'indices': [
                    # ..:    {X ... }Y, ...
                    # ....]
                    # .}
                    {
                        "$project": dict.fromkeys(
                            (
                                "build_config",
                                "environment",
                                "created_at",
                            ),
                            1,
                        )
                    },
                    {"$sort": {"created_at": 1}},
                    {
                        "$group": {
                            "_id": "$build_config",
                            "indices": {"$push": "$$ROOT"},
                        }
                    },
                ]
            )
        )  # ............................: }
        return _CleanUps(
            [
                _BuildConfig(
                    doc["_id"],  # ↓ -0 in slicing does not yield the desired result
                    _IndicesToRemove(doc["indices"][: -keep or len(doc["indices"])]),
                    _IndicesToKeep(doc["indices"][-keep or len(doc["indices"]) :]),
                )
                for doc in results
            ]
        )

    async def clean(self, cleanups):
        self.logger.debug(cleanups)

        actions = CleanUpResult()
        for index in itertools.chain.from_iterable(cleanups):
            args = self.indexers[index["environment"]]["args"]

            async with AsyncElasticsearch(**args) as client:
                await client.indices.delete(index["_id"], ignore_unavailable=True)

                action = ("DELETE", str(index))
                actions.append(action)
                logging.info(action)

                self.collection.update_many(
                    {f"index.{index['_id']}.environment": index["environment"]},
                    {"$unset": {f"index.{index['_id']}": 1}},
                )
        return actions

    @staticmethod
    def plain_text(cleanups):
        plain_texts = []
        for build_config in cleanups:
            plain_texts.append(f"> BuildConfig {repr(build_config.name)}")
            plain_texts.append(f"     Found {len(build_config.remove)} indices to remove:")
            for index in build_config.remove:
                _id = index.get("_id")
                env = repr(index.get("environment"))
                ts = str(index.get("created_at"))
                plain_texts.append(f"        {_id} (env={env}, created={ts})")
            plain_texts.append(f"     Found {len(build_config.keep)} indices to keep:")
            for index in build_config.keep:
                _id = index.get("_id")
                env = repr(index.get("environment"))
                ts = str(index.get("created_at"))
                plain_texts.append(f"        {_id} (env={env}, created={ts})")
            plain_texts.append("")
        return "\n".join(plain_texts)


# Feature Specification ↑
# https://suwulab.slack.com/archives/CC19LHAF2/p1631119811009900?thread_ts=1631063230.003400&cid=CC19LHAF2

# >BuildConfig "mygene_allspecies":
#     Found 8 indices to remove:
#        mygene_allspecies_20180501_rpds7zzn (indexer_env='prod', created=2018-05-01T04:45:25.132000)
#        ...

#     Found 3 indices to keep:
#        mygene_allspecies_20180501_rpds7zzn (indexer_env='prod', created=2021-04-06T22:34:09.863000)

# NOTE
# Using "env" instead of "indexer_env" as desribed in the feature spec
# to match the signature of the Cleaner.find method (first argument).


def test_str():
    print(
        _CleanUps(
            [
                _BuildConfig(
                    "mynews",
                    _IndicesToRemove(
                        [
                            {
                                "_id": "mynews_20210811_test",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2021, 8, 11, 19, 27, 25, 141000),
                                "environment": "local",
                            },
                            {
                                "_id": "mynews_202105261855_5ffxvchx",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2021, 8, 16, 9, 26, 56, 221000),
                                "environment": "local",
                            },
                            {
                                "_id": "mynews_202012280220_vsdevjdk",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2021, 8, 17, 0, 23, 11, 374000),
                                "environment": "local",
                            },
                        ]
                    ),
                    _IndicesToKeep(
                        [
                            {
                                "_id": "mynews_202009170234_fjvg7skx",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2020, 9, 17, 2, 35, 12, 800000),
                                "environment": "local",
                            },
                            {
                                "_id": "mynews_202009222133_6rz3vljq",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2020, 9, 22, 21, 33, 40, 958000),
                                "environment": "local",
                            },
                            {
                                "_id": "mynews_202010060100_ontyofuv",
                                "build_config": "mynews",
                                "created_at": datetime.datetime(2020, 10, 6, 1, 0, 11, 237000),
                                "environment": "local",
                            },
                        ]
                    ),
                )
            ]
        )
    )


def test_find():
    from pymongo import MongoClient

    logging.basicConfig(level="DEBUG")

    # mychem
    # -------
    # "su04"
    # "mychem_hubdb", "src_build"

    # mygene
    # -------
    # "su05"
    # "genedoc_src", "src_build"

    # docker
    # ------
    #
    # "biothings", "src_build"

    client = MongoClient("su05")
    collection = client["genedoc_src"]["src_build"]
    cleaner = Cleaner(collection, {"local": {"args": {}}})
    obj = cleaner.find()
    # print(type(obj))
    # print(obj)
    print(cleaner.plain_text(obj))
    return cleaner, obj


def test_clean():
    import asyncio

    cleaner, cleanups = test_find()
    loop = asyncio.get_event_loop()
    print(loop.run_until_complete(cleaner.clean(cleanups)))


if __name__ == "__main__":
    test_find()



# DatabaseCollectionTesting

# Author: Amiteshk Sharma
# https://github.com/amiteshksharma/Database-Testing

import pymongo

# from jsonpath_ng import jsonpath, parse


class DatabaseCollectionTesting:

    """
    Constructor that takes in three items
    db_url      - string - the mongoDB url to connect to
    db          - string - name of DB to use
    collection  - string - name of collection in db
    """

    def __init__(self, db_url, db, collection):
        if "mongo" not in db_url:
            raise ValueError

        self.database_url = db_url
        self.client = pymongo.MongoClient(db_url)
        self.db = self.client[db]
        self.collection = self.db[collection]

    # test to see if there exists only 1 item for an ID
    # _id - the document _id to query
    def test_field_unique_id(self, _id):
        get_item = self.collection.find({"_id": _id})
        item = list(get_item)
        assert len(item) == 1

    # check for all items within a taxid
    # taxid - the taxid of the document to query
    def test_field_taxid(self, taxid):
        get_items = self.collection.find({"taxid": taxid})
        item_list = list(get_items)
        assert len(item_list) >= 1

    # check all documents with the same taxid value
    def test_documents_taxid(self, taxid):
        get_documents = self.collection.find({"taxid": taxid})
        doc_list = list(get_documents)

        sub_list = ["_id", "taxid", "name", "ensembl", "symbol"]

        for doc in doc_list:
            keys = doc.keys()
            if not all(x in keys for x in sub_list):
                # may not contain a name attribute
                if "name" not in keys:
                    pass
                # may not contain the ensembl attribute
                elif "ensembl" not in keys:
                    pass
                # may not contain the symbol attribute
                elif "symbol" not in keys:
                    pass
                else:
                    # assert False
                    raise AssertionError()

        assert True

    # check an _id and make sure it does not exist
    # _id - the document _id to query
    def test_field_does_not_exist(self, _id):
        get_items = self.collection.find({"_id": _id})
        item_list = list(get_items)
        assert len(item_list) == 0

    # Check number of documents is correct
    # expected count - the expected count of documents with specific taxid
    def test_total_document_count(self, expected_count):
        get_all_document = self.collection.find()
        document_list = list(get_all_document)
        assert len(document_list) == expected_count

    # check the indices for the mongoDB database
    def test_database_index(self):
        get_indices = self.collection.index_information()
        indices_list = list(get_indices)
        size = len(indices_list)

        # if only size 1, then it only has _id_ index
        if size == 1:
            assert all(x in indices_list for x in ["_id_"])
        elif size == 3:
            assert all(x in indices_list for x in ["_id_", "taxid_1", "entrezgene_1"])

    # test the name attribute on randomly selected items in the database
    def test_document_name(self):
        random_docs = self.collection.aggregate([{"$sample": {"size": 10}}])

        count = 0
        for doc in random_docs:
            if "name" in doc:
                count = count + 1
            else:
                print("_id for document with no name: " + doc["_id"])

        assert count == 10


if __name__ == "__main__":
    c = DatabaseCollectionTesting("mongodb://su05:27017", "genedoc", "mygene_allspecies_20191111_eeesndlz")
    c.test_documents_taxid(29302)



import asyncio
import datetime
import json
import os
from functools import partial
from typing import Optional

from elasticsearch import Elasticsearch, NotFoundError, RequestsHttpConnection
from requests_aws4auth import AWS4Auth

import biothings.hub.dataload.uploader as uploader
from biothings import config as btconfig
from biothings.utils.backend import DocESBackend
from biothings.utils.common import get_random_string
from biothings.utils.es import IndexerException


class BiothingsUploader(uploader.BaseSourceUploader):
    name = None

    # Specify the backend this uploader should work with. Must be defined before instantiation
    # (can be an instance or a partial() returning an instance)
    TARGET_BACKEND = None

    # Specify the syncer function this uploader will use to apply diff
    # (can be an instance or a partial() returning an instance)
    SYNCER_FUNC = None

    # should we delete index before restoring snapshot if index already exist ?
    AUTO_PURGE_INDEX = False

    def __init__(self, *args, **kwargs):
        super(BiothingsUploader, self).__init__(*args, **kwargs)
        self._target_backend = None
        self._syncer_func = None

    @property
    def target_backend(self):
        if not self._target_backend:
            if type(self.__class__.TARGET_BACKEND) == partial:
                self._target_backend = self.__class__.TARGET_BACKEND()
            else:
                self._target_backend = self.__class__.TARGET_BACKEND
            assert (
                type(self._target_backend) == DocESBackend
            ), "Only ElasticSearch backend is supported (got %s)" % type(self._target_backend)
        return self._target_backend

    @property
    def syncer_func(self):
        if not self._syncer_func:
            self._syncer_func = self.__class__.SYNCER_FUNC
        return self._syncer_func

    async def load(self, *args, **kwargs):
        return await super().load(steps=["data"], *args, **kwargs)

    async def update_data(self, batch_size, job_manager, **kwargs):
        """
        Look in data_folder and either restore a snapshot to ES
        or apply diff to current ES index
        """
        # determine if it's about a snapshot/full and diff/incremental
        # we should have a json metadata matching the release
        self.prepare_src_dump()  # load infor from src_dump
        release = self.src_doc.get("download", {}).get("release")
        assert release, "Can't find release information in src_dump document"
        build_meta = json.load(open(os.path.join(self.data_folder, "%s.json" % release)))
        if build_meta["type"] == "full":
            res = await self.restore_snapshot(build_meta, job_manager=job_manager, **kwargs)
        elif build_meta["type"] == "incremental":
            res = await self.apply_diff(build_meta, job_manager=job_manager, **kwargs)
        return res

    def get_snapshot_repository_config(self, build_meta):
        """Return (name,config) tuple from build_meta, where
        name is the repo name, and config is the repo config"""
        # repo_name, repo_settings = list(
        #     build_meta["metadata"]["repository"].items())[0]
        # TODO
        repo_name = build_meta["metadata"]["repository"]["name"]
        repo_settings = build_meta["metadata"]["repository"]
        return (repo_name, repo_settings)

    def _get_es_client(self, es_host: str, auth: Optional[dict]):
        """
        Get Elasticsearch Client

        Used by self._get_repository, self._create_repository
        """
        es_conf = {
            "timeout": 120,
            "max_retries": 3,
            "retry_on_timeout": False,
        }
        if auth:
            # see https://git.io/JoAE4 on BioThings.API Wiki
            if auth["type"] == "aws":
                auth_args = (
                    auth["properties"]["access_id"],
                    auth["properties"]["secret_key"],
                    auth["properties"]["region"],
                    "es",
                )
                es_conf["http_auth"] = AWS4Auth(*auth_args)
                es_conf["connection_class"] = RequestsHttpConnection
            elif auth["type"] == "http":
                auth_args = (
                    auth["properties"]["username"],
                    auth["properties"]["password"],
                )
                es_conf["http_auth"] = auth_args
            else:
                raise RuntimeError("Auth settings not recognized")
        es = Elasticsearch(es_host, **es_conf)
        return es

    def _get_repository(self, es_host: str, repo_name: str, auth: Optional[dict]):
        es = self._get_es_client(es_host, auth)
        try:
            repo = es.snapshot.get_repository(repository=repo_name)
        except NotFoundError:
            repo = None
        return repo

    def _create_repository(self, es_host: str, repo_name: str, repo_settings: dict, auth: Optional[dict]):
        """
        Create Elasticsearch Snapshot repository
        """
        es = self._get_es_client(es_host, auth)
        es.snapshot.create_repository(repository=repo_name, body=repo_settings)

    async def restore_snapshot(self, build_meta, job_manager, **kwargs):
        self.logger.debug("Restoring snapshot...")
        idxr = self.target_backend.target_esidxer
        es_host = idxr.es_host
        self.logger.debug("Got ES Host: %s", es_host)
        repo_name, repo_settings = self.get_snapshot_repository_config(build_meta)
        self.logger.debug("Got repo name: %s", repo_name)
        self.logger.debug("With settings: %s", repo_settings)
        # pull authentication settings from config
        auth = btconfig.STANDALONE_CONFIG.get(self.name, {}).get(
            "auth", btconfig.STANDALONE_CONFIG["_default"].get("auth")
        )
        if auth:
            self.logger.debug("Obtained Auth settings, using them.")
        else:
            self.logger.debug("No Auth settings found")

        # all restore repos should be r/o
        repo_settings["settings"]["readonly"] = True

        # populate additional settings
        additional_settings = btconfig.STANDALONE_CONFIG.get(self.name, {}).get(
            "repo_settings", btconfig.STANDALONE_CONFIG["_default"].get("repo_settings")
        )
        if additional_settings:
            self.logger.debug("Adding additional settings: %s", additional_settings)
            repo_settings["settings"].update(additional_settings)

        if "client" not in repo_settings["settings"]:
            self.logger.warning("\"client\" not set in repository settings. The 'default' " "client will be used.")
            self.logger.warning(
                "Make sure keys are in the Elasticsearch keystore. "
                "If you are trying to work with EOL versions of "
                "Elasticsearch, or if you intentionally enabled "
                'allow_insecure_settings, set "access_key", "secret_key",'
                " and potentially \"region\" in additional 'repo_settings'."
            )

        # first check if snapshot repo exists
        self.logger.info("Getting current repository settings")
        existing_repo_settings = self._get_repository(es_host, repo_name, auth)
        if existing_repo_settings:
            if existing_repo_settings[repo_name] != repo_settings:
                # TODO update comparison logic
                self.logger.info(
                    f"Repository '{repo_name}' was found but settings are different, " "it may need to be created again"
                )
                self.logger.debug("Existing setting: %s", existing_repo_settings[repo_name])
                self.logger.debug("Required (new) setting: %s" % repo_settings)
            else:
                self.logger.info("Repo exists with correct settings")
        else:
            # ok, it doesn't exist let's try to create it
            self.logger.info("Repo does not exist")
            try:
                self.logger.info("Creating repo...")
                self._create_repository(es_host, repo_name, repo_settings, auth)
            except Exception as e:
                self.logger.info("Creation failed: %s", e)
                if "url" in repo_settings["settings"]:
                    raise uploader.ResourceError(
                        "Could not create snapshot repository. Check elasticsearch.yml configuration "
                        + "file, you should have a line like this: "
                        + 'repositories.url.allowed_urls: "%s*" ' % repo_settings["settings"]["url"]
                        + "allowing snapshot to be restored from this URL. Error was: %s" % e
                    )
                else:
                    raise uploader.ResourceError("Could not create snapshot repository: %s" % e)

        # repository is now ready, let's trigger the restore
        snapshot_name = build_meta["metadata"]["snapshot_name"]

        # backup the original value of indexer's replica
        original_number_of_replicas = idxr.get_internal_number_of_replicas() or 0
        alias_name = idxr.canonical_index_name

        use_no_downtime_method = kwargs.get("use_no_downtime_method", True)
        append_ts = kwargs.get("append_ts", True)
        if use_no_downtime_method:
            base_index_name = snapshot_name
            if append_ts:
                ts = datetime.datetime.now().strftime("%Y%m%d%H%M")
                base_index_name = f"{snapshot_name}_{ts}"
            if len(base_index_name) >= 255:
                raise RuntimeError("Deterministic part of index name already too long")

            index_name = base_index_name
            append_random_str = False
            while True:
                if append_random_str:
                    index_name += "_" + get_random_string()
                    index_name = index_name[:255]  # elasticsearch restriction
                if not idxr.exists_index(index=index_name):
                    break
        else:
            index_name = alias_name

        pinfo = self.get_pinfo()
        pinfo["step"] = "restore"
        pinfo["description"] = snapshot_name

        def get_status_info():
            try:
                res = idxr.get_restore_status(index_name)
                return res
            except Exception as e:
                # somethng went wrong, report as failure
                return {"status": "FAILED %s" % e}

        def done_callback(f, step: str):
            try:
                self.logger.info("%s launched: %s" % (step, f.result()))
            except Exception as e:
                self.logger.error("Error while launching %s: %s" % (step, e))
                raise e

        self.logger.info(
            "Restoring snapshot '%s' to index '%s' on host '%s'" % (snapshot_name, index_name, idxr.es_host)
        )
        # ESIndexer.restore is synchronous but should return relatively
        # quickly
        job = await job_manager.defer_to_thread(
            pinfo, partial(idxr.restore, repo_name, snapshot_name, index_name, purge=self.__class__.AUTO_PURGE_INDEX)
        )
        job.add_done_callback(partial(done_callback, step="restore"))
        await job

        def update_alias_and_delete_old_indices():
            # Find indices which starts with snapshot_name, and sort by creation date and order by asc
            old_indices = []
            try:
                old_indices.extend(
                    idxr.get_indice_names_by_settings(index=alias_name + "*", sort_by_creation_date=True, reverse=False)
                )
            except Exception:
                pass
            self.logger.debug("Alias '%s' points to '%s'" % (alias_name, old_indices))
            if index_name in old_indices:
                self.logger.warning("new index name in old alias, something is not right")
                self.logger.warning("continuing alias swap despite potential problem")
                old_indices.remove(index_name)
            try:
                idxr.update_alias(alias_name, index_name)
                self.logger.info(f"Alias '{alias_name}' updated to " f"associate with index '{index_name}'")
            except IndexerException as e:
                self.logger.warning(f"Alias index swap ran into a problem {e}")
                self.logger.warning(f"Deleting new index '{index_name}'")
                idxr.delete_index(index_name)
                raise

            # have ESIndexer look at the correct index after snapshot restore
            idxr.check_index()

            # after successful swap, delete old indices
            # only issue messages on errors
            # we only keep n recent indices depends on the config.RELEASE_KEEP_N_RECENT_INDICES
            # n < 0: keep all
            # n == 0: only keep the new created indice
            # n > 0: only keep at most n latest indices
            number_indexes_to_keep = btconfig.RELEASE_KEEP_N_RECENT_INDICES
            if number_indexes_to_keep > 0:
                old_indices = old_indices[:-number_indexes_to_keep]
            elif number_indexes_to_keep < 0:
                old_indices = []

            try:
                for rm_idx_name in old_indices:
                    idxr.delete_index(rm_idx_name)
                    self.logger.info("Deleted old index '%s'" % rm_idx_name)
            except Exception:  # nosec
                # just inform the user that deletion failed, not that harmful
                self.logger.error("Failed to delete old indices, try deleting " f"{old_indices} manually")

            # restore indexer's replica to original value
            idxr.set_internal_number_of_replicas(original_number_of_replicas)

        while True:
            status_info = get_status_info()
            status = status_info["status"]
            self.logger.info("Recovery status for index '%s': %s" % (index_name, status_info))
            if status in ["INIT", "IN_PROGRESS"]:
                await asyncio.sleep(getattr(btconfig, "MONITOR_SNAPSHOT_DELAY", 60))
            else:
                if status == "DONE":
                    self.logger.info(
                        "Snapshot '%s' successfully restored to index '%s' (host: '%s')"
                        % (snapshot_name, index_name, idxr.es_host),
                        extra={"notify": True},
                    )
                    if use_no_downtime_method:
                        job = await job_manager.defer_to_thread(pinfo={}, func=update_alias_and_delete_old_indices)
                        job.add_done_callback(partial(done_callback, step="alias"))
                        await job
                else:
                    e = uploader.ResourceError(
                        "Failed to restore snapshot '%s' on index '%s', status: %s"
                        % (snapshot_name, idxr._index, status)
                    )
                    self.logger.error(e)
                    raise e
                break
        # return current number of docs in index
        return self.target_backend.count()

    async def apply_diff(self, build_meta, job_manager, **kwargs):
        self.logger.info("Applying incremental update from diff folder: %s" % self.data_folder)
        meta = json.load(open(os.path.join(self.data_folder, "metadata.json")))
        # old: index we want to update
        old = (
            self.target_backend.target_esidxer.es_host,
            meta["old"]["backend"],
            # TODO
            # target name can be release index name,
            # maybe should refer to old backend name
            # ----------------------------------------
            #  self.target_backend.target_name,
            # ----------------------------------------
            self.target_backend.target_esidxer._doc_type,
        )
        # new: index's data we will reach once updated (just informative)
        new = (
            self.target_backend.target_esidxer.es_host,
            meta["new"]["backend"],
            self.target_backend.target_esidxer._doc_type,
        )
        await self.syncer_func(old_db_col_names=old, new_db_col_names=new, diff_folder=self.data_folder)
        # return current number of docs in index (even if diff update)
        return self.target_backend.count()

    def clean_archived_collections(self):
        pass



import logging
import time
from typing import Optional, Union

from biothings.utils.es import ESIndexer

logger = logging.getLogger(__name__)


def reindex(
    src_index: str,
    target_index: Optional[str] = None,
    settings: Optional[dict] = None,
    mappings: Optional[dict] = None,
    alias: Optional[Union[bool, str]] = None,
    delete_src: Optional[bool] = False,
    task_check_interval: Optional[int] = 15,
    src_indexer_kwargs: Optional[dict] = None,
    target_indexer_kwargs: Optional[dict] = None,
    reindex_kwargs: Optional[dict] = None,
) -> None:
    """
    This helper function helps to reindex an existing index by transferring both the settings, mappings and docs.
    Mappings and settings from the src index will be used to create a new empty target index first.
    Then Elasticsearch's reindex API will be used to transfer all docs to the new one.
    Optionally, the alias can be switched over to the new index too.
    This is useful when we need to migrate existing indices created from the srcer ES version to the current ES version.

    Parameters:
        src_index: name of the src index

        target_index: name of the new index, use <src_index_name>_reindexed as default if None

        settings: if provided as a dict, update the settings with the provided dictionary.
                    Otherwise, keep the same from the src_index
        mapping: if provided as a dict, update the settings with the provided mappings.
                    Otherwise, keep the same from the src_index
        alias: if True, switch the alias from src_index to target_index.
                If src_index has no alias, apply the <src_index_name> as the alias;
                if a string value, apply it as the alias instead
        delete_src: If True, delete the src_index after everything is done,
        task_check_interval: the interval to check reindex job status, default is 15s
        src_indexer_kwargs: a dict contains infor to construct ESIndexer for src index
        target_indexer_kwargs: a dict contains infor to construct ESIndexer for target index
        reindex_kwargs: a dict contains extra parameters passed to reindex, e.g. {"timeout": "5m",
                        "slices": 10}. see
        https://www.elastic.co/guide/en/elasticsearch/reference/7.17/docs-reindex.html#docs-reindex-api-query-params
                        for available parameters. Note that "wait_for_completion" parameter
                        will always be set to False, so we will run reindex asynchronously

    Returns:
        None
    """

    logger.info("Starting reindex...")

    # create indexer objects
    logger.info("Create src index obj")
    src_indexer_kwargs = src_indexer_kwargs or {}
    src_index_obj = ESIndexer(index=src_index, **src_indexer_kwargs)
    assert src_index_obj.exists_index(), f"src index '{src_index}' does not exists."
    # clone settings, mappings from src index
    logger.info("Clone src settings, and src_mapping, and update with supplied values")
    src_settings = src_index_obj.get_settings(src_index_obj._index) or {}
    src_mappings = src_index_obj.get_mapping() or {}
    src_index_obj.number_of_shards = src_settings.get("number_of_shards", 1)
    src_index_obj.number_of_replicas = src_settings.get("number_of_replicas", 0)

    # update src settings, mappings with supplied values.
    if settings:
        src_settings.update(**settings)
    if mappings:
        src_mappings.update(**mappings)

    logger.info("Create target index obj")
    target_index = target_index or f"{src_index}_reindexed"
    target_indexer_kwargs = target_indexer_kwargs or {}
    if "number_of_shards" not in target_indexer_kwargs:
        target_indexer_kwargs["number_of_shards"] = src_index_obj.number_of_shards
    if "number_of_replicas" not in target_indexer_kwargs:
        target_indexer_kwargs["number_of_replicas"] = src_index_obj.number_of_replicas
    target_index_obj = ESIndexer(index=target_index, **target_indexer_kwargs)
    assert not target_index_obj.exists_index(), f"target index '{target_index}' exists."

    # create target index with src settings, mappings
    logger.info("Set target index's settings, mappings")
    target_index_obj.create_index(mapping=src_mappings, extra_settings=src_settings[src_index_obj._index]["settings"])

    # Reindex from src index to target index
    logger.info("Reindex from src index to target index")
    is_remote = src_index_obj.es_host != target_index_obj.es_host
    reindex_kwargs = reindex_kwargs or {}
    # always run reindex asynchronously, otherwise reindex method below tends to be timeout eventually
    reindex_kwargs["wait_for_completion"] = False
    result = target_index_obj.reindex(src_index_obj, is_remote=is_remote, **reindex_kwargs)
    # result: {'task': '4ohWVeJGQVq5lZCNJgkPag:149689'}
    task_id = result.get("task")
    logger.info('Reindex job started [task id: "%s"]:', task_id)
    # check task status till it's completed
    while 1:
        task_obj = target_index_obj._es.tasks.get(task_id)
        task_status = task_obj.get("task", {}).get("status", {})
        logger.info(
            "\t[Progress] completed: %s, batches: %s, created: %s",
            task_obj.get("completed"),
            task_status.get("batches"),
            task_status.get("created"),
        )
        if task_obj.get("completed", False) is True:
            break
        time.sleep(task_check_interval)

    # Refresh and flush target index
    logger.info("Flush and refresh target index")
    target_index_obj.flush_and_refresh()

    # Check doc counts to make sure src_index, new index are equals
    if "max_docs" in reindex_kwargs:
        # if max_docs is set, the target_index will only contain that number of docs
        count_src_docs = reindex_kwargs["max_docs"]
    else:
        count_src_docs = src_index_obj.count()
    count_target_docs = target_index_obj.count()
    assert (
        count_src_docs == count_target_docs
    ), "Number of docs in target index not equal to number of docs in src index"
    logger.info("Verified number of docs in target index is equal to number of docs in src index")

    # Set target index's alias
    if alias:
        logger.info("Update target index's alias")
        if isinstance(alias, str):
            new_alias = alias
        else:
            try:
                alias_data = src_index_obj.get_alias(index=src_index_obj._index)
                new_alias = list(alias_data[src_index_obj._index]["aliases"].keys())[0]
            except:  # noqa
                # default to src_index name if src index is deleted or has no alias.
                new_alias = src_index_obj._index
        target_index_obj.update_alias(new_alias)

    # Delete src index
    if delete_src:
        logger.info("Delete src alias")
        src_index_obj.delete_index()

    logger.info("Finished!")



import urllib.parse
from typing import Any, Union

from tornado.auth import OAuth2Mixin
from tornado.escape import json_decode
from tornado.httputil import HTTPHeaders

__all__ = [
    "GithubOAuth2Mixin",
    "OrcidOAuth2Mixin",
]


class GithubOAuth2Mixin(OAuth2Mixin):
    """
    Mixin Class for using GitHub API with OAuth2

    Currently only two methods are implemented:
     - Getting OAuth2 Token using an authorization code for exchange
       (Authorization Code Flow, aka 3-legged OAuth)
     - Reading user information given token

    TODO: if anyone adds any new features in the future, make a dedicated
          OAuth2 request method, similar to the one for ORCID below.
    """

    _OAUTH_AUTHORIZE_URL = "https://github.com/login/oauth/authorize"
    _OAUTH_ACCESS_TOKEN_URL = "https://github.com/login/oauth/access_token"

    _GITHUB_API_URL_BASE = "https://api.github.com/"

    async def github_get_oauth2_token(
        self,
        client_id: str,
        client_secret: str,
        code: str,
    ) -> dict:
        """
        Get Github OAuth2 Token

        Returns:
            Dictionary with keys access_token, scope, token_type. See https://git.io/J1ON4

        Raises:
            HTTPError: if request fails
            JSONDecodeError: if parsing the response fails
        """
        http = self.get_auth_http_client()
        args = {
            "code": code,
            "client_id": client_id,
            "client_secret": client_secret,
        }
        headers = HTTPHeaders()
        headers.add("Accept", "application/json")
        response = await http.fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            raise_error=True,
            method="POST",
            body=urllib.parse.urlencode(args),
            headers=headers,
        )
        ret = json_decode(response.body)
        return ret

    async def github_get_authenticated_user(self, token: Union[dict, str]) -> dict:
        """
        Get Github User Info

        Basically, only request the /user endpoint with the token provided.
        The token does not need special scopes to read basic user information
        (like Github username or user ID), but to get the user email in the
        response, a token with 'user:email' scope is needed.

        Caveats: unlike the torngithub package mixin or the Facebook mixin,
        this method has a different signature and accepts different parameters
        (just the token). Also unlike the Google mixin this actually returns the
        user, not the token. So, to replace the mixin, one method call becomes
        two calls, first get the token, then get the user.

        Another caveat is that the callback functions have been removed. We
        weren't using them anyways, and sane humans don't like callback hell.

        Args:
            token: token string or raw token dict obtained from get_oauth2_token in
            this mixin class

        Returns:
            Github user information in a dict. What you get depends on the scope
            of the token provided.

        Raises:
            HTTPError: if request fails
            JSONDecodeError: if parsing the response fails
            RuntimeError: when using an invalid token or unsupported token type

        Example:
            >>> token = await self.github_get_oauth2_token(
            >>>     'SOME_CLIENT_ID', 'SOME_CLIENT_SECRET',
            >>>     'THE_CODE_GITHUB_PROVIDED_IN_THE_REDIRECT'
            >>> )
            >>> user = await self.github_get_authenticated_user(token)
            >>> print(user['login'], user['id'])
            zcqian 7196478

        """
        http = self.get_auth_http_client()
        headers = HTTPHeaders()
        if isinstance(token, str):
            token_str = token
        elif isinstance(token, dict):
            if "token_type" not in token or token["token_type"] != "bearer" or "access_token" not in token:
                raise RuntimeError("Token seems invalid")
            else:
                token_str = token["access_token"]
        else:
            raise RuntimeError("Token seems invalid")
        headers.add("Authorization", f"token {token_str}")
        headers.add("Accept", "application/vnd.github.v3+json")
        resp = await http.fetch(urllib.parse.urljoin(self._GITHUB_API_URL_BASE, "user"), method="GET", headers=headers)
        ret = json_decode(resp.body)
        return ret


class OrcidOAuth2Mixin(OAuth2Mixin):
    """
    Mixin Class for using ORCID API with OAuth2

    Note:
        When redirecting the user to the authorization page, do NOT use
        the '/read-public' scope. Either get it using OAuth2 client
        credentials flow or just use '/authenticate' or 'openid' scopes,
        the token returned with these two scopes can be used to read
        public data.
    """

    _OAUTH_AUTHORIZE_URL = "https://orcid.org/oauth/authorize"
    _OAUTH_ACCESS_TOKEN_URL = "https://orcid.org/oauth/token"
    _ORCID_API_URL_BASE = "https://pub.orcid.org/v2.0/"

    async def orcid_get_oauth2_token(
        self,
        client_id: str,
        client_secret: str,
        code: str,
    ) -> dict:
        """
        Get OAuth2 Token from ORCID

        See OAuth2 Authorization Code Flow.

        Returns:
            Dictionary with access_token. If `openid` scope is used, key 'id_token'
                will also be present.

        Raises:
            HTTPError: if request fails
            JSONDecodeError: if parsing the response fails
        """
        http = self.get_auth_http_client()
        args = {
            "code": code,
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "authorization_code",
        }
        headers = HTTPHeaders()
        headers.add("Accept", "application/json")
        response = await http.fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            raise_error=True,
            method="POST",
            body=urllib.parse.urlencode(args),
            headers=headers,
        )
        ret = json_decode(response.body)
        return ret

    async def orcid_oauth2_request(
        self,
        url: str,
        access_token: Union[dict, str],
        method: str = "GET",
        **kwargs: Any,
    ):
        """
        Make Request to ORCID API With OAuth2 Token

        Args:
            url: The full request URL in string
            access_token: the ORCID access token, either the entire object or the token
                in string. It is NOT the OpenID Connect ID Token.
            method: HTTP Method to use, string.
            **kwargs: Any additional keyword arguments are directly passed to the
                tornado.httpclient.AsyncHTTPClient.fetch method.
        Raises:
            ValueError: if the token looks obviously invalid
            HTTPError: if the request fails. Note Tornado seems to use 599 for timeouts
            JSONDecodeError: if JSON decoding fails. Check if you're making requests
                to the correct ORCID endpoint.
        """
        http = self.get_auth_http_client()
        headers = HTTPHeaders()
        if isinstance(access_token, str):
            token_str = access_token
        elif isinstance(access_token, dict):
            if (
                "token_type" not in access_token
                or access_token["token_type"] != "bearer"
                or "access_token" not in access_token
            ):
                raise ValueError("Token seems invalid")
            else:
                token_str = access_token["access_token"]
        else:
            raise ValueError("Token seems invalid")
        headers.add("Authorization", f"Bearer {token_str}")
        headers.add("Accept", "application/json")
        resp = await http.fetch(url, method=method, headers=headers, **kwargs)
        ret = json_decode(resp.body)
        return ret

    async def orcid_get_authenticated_user_oidc(self, access_token: Union[dict, str]) -> dict:
        """
        Get ORCID User OpenID Connect information

        See ORCID Documentation https://git.io/JyY23 or its latest version.

        ORCID supports OpenID Connect and this provides the OpenID Connect user details
        format. This method obtains such information.

        Args:
             access_token: the ORCID access token, either the entire object or the token
                in string. It is NOT the OpenID Connect ID Token.

        Returns:
            Dictionary with OIDC User Details
        """
        ret = await self.orcid_oauth2_request(
            urllib.parse.urljoin(self._OAUTH_ACCESS_TOKEN_URL, "userinfo"), access_token, "GET"
        )
        return ret

    async def orcid_get_authenticated_user_record(
        self,
        access_token: Union[dict, str],
        orcid: str,
    ) -> dict:
        """
        Get ORCID User Record

        Get the full ORCID user record. Note, what you get depends on what the user
        allows you to read.

        Args:
            access_token: the ORCID access token, either the entire object or the token
                in string. It is NOT the OpenID Connect ID Token.
            orcid: the ORCID of the authenticated user. Typically you can find this
                in the token object under the 'orcid' field.
        Returns:
            Dictionary with ORCID Record
        """
        ret = await self.orcid_oauth2_request(
            urllib.parse.urljoin(self._ORCID_API_URL_BASE, f"{orcid}/record"), access_token, "GET"
        )
        return ret



"""
    Search Execution Engine

    Take the output of the query builder and feed
    to the corresponding database engine. This stage
    typically resolves the db destination from a
    biothing_type and applies presentation and/or
    networking parameters.

    Example:

    >>> from biothings.web.query import ESQueryBackend
    >>> from elasticsearch import Elasticsearch
    >>> from elasticsearch_dsl import Search

    >>> backend = ESQueryBackend(Elasticsearch())
    >>> backend.execute(Search().query("match", _id="1017"))

    >>> _["hits"]["hits"][0]["_source"].keys()
    dict_keys(['taxid', 'symbol', 'name', ... ])

"""

import asyncio

from elasticsearch import NotFoundError, RequestError
from elasticsearch_dsl import MultiSearch, Search

from biothings.web.query.builder import ESScrollID


class ResultInterrupt(Exception):
    def __init__(self, data):
        super().__init__()
        self.data = data


class RawResultInterrupt(ResultInterrupt):
    pass


class EndScrollInterrupt(ResultInterrupt):
    def __init__(self):
        super().__init__({"success": False, "error": "No more results to return."})


class ESQueryBackend:
    def __init__(self, client, indices=None):
        self.client = client
        self.indices = indices or {None: "_all"}

        # a list of biothing_type -> index pattern mapping
        # ---------------------------------------------------
        # {
        #   None: "hg19_current",
        #   "hg19": "hg19_current",
        #   "hg38": "hg38_index1,hg38_index2",
        #   "_internal": "hg*_current"
        # }

        if None not in self.indices:  # set default index pattern
            self.indices[None] = next(iter(self.indices.values()))

    def execute(self, query, **options):
        assert isinstance(query, Search)
        index = self.indices[options.get("biothing_type")]
        # index can be further adjusted (e.g. based on options) if necessary
        index = self.adjust_index(index, query, **options)
        return self.client.search(query.to_dict(), index)

    def adjust_index(self, original_index, query, **options):
        """
        Override to get specific ES index.
        """
        return original_index


class AsyncESQueryBackend(ESQueryBackend):
    """
    Execute an Elasticsearch query
    """

    def __init__(
        self,
        client,
        indices=None,
        scroll_time="1m",
        scroll_size=1000,
        multisearch_concurrency=5,
        total_hits_as_int=True,
    ):
        super().__init__(client, indices)

        # for scroll queries
        self.scroll_time = scroll_time  # scroll context expiration timeout
        self.scroll_size = scroll_size  # result window size override value

        # concurrency control
        self.semaphore = asyncio.Semaphore(multisearch_concurrency)

        # additional params
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/breaking-changes-7.0.html
        # #hits-total-now-object-search-response
        self.total_hits_as_int = total_hits_as_int

    async def execute(self, query, **options):
        """
        Execute the corresponding query. Must return an awaitable.
        May override to add more. Handle uncaught exceptions.

        Options:
            fetch_all: also return a scroll_id for this query (default: false)
            biothing_type: which type's corresponding indices to query (default in config.py)
        """
        assert isinstance(
            query,
            (
                # https://www.elastic.co/guide/en/elasticsearch/reference/current/search-search.html
                # https://www.elastic.co/guide/en/elasticsearch/reference/current/search-multi-search.html
                # https://www.elastic.co/guide/en/elasticsearch/reference/current/scroll-api.html
                Search,
                MultiSearch,
                ESScrollID,
            ),
        )

        if isinstance(query, ESScrollID):
            try:
                res = await self.client.scroll(
                    scroll_id=query.data, scroll=self.scroll_time, rest_total_hits_as_int=self.total_hits_as_int
                )
            except (
                RequestError,  # the id is not in the correct format of a context id
                NotFoundError,  # the id does not correspond to any search context
            ):
                raise ValueError("Invalid or stale scroll_id.")
            else:
                if options.get("raw"):
                    raise RawResultInterrupt(res)

                if not res["hits"]["hits"]:
                    raise EndScrollInterrupt()

                return res

        # everything below require us to know which indices to query
        index = self.indices[options.get("biothing_type")]
        # index can be further adjusted (e.g. based on options) if necessary
        index = self.adjust_index(index, query, **options)

        if isinstance(query, Search):
            if options.get("fetch_all"):
                query = query.extra(size=self.scroll_size)
                query = query.params(scroll=self.scroll_time)
            if self.total_hits_as_int:
                query = query.params(rest_total_hits_as_int=True)
            query_kwargs = query.to_dict()
            query_kwargs.update(query._params)
            if "from" in query_kwargs:
                query_kwargs["from_"] = query_kwargs.pop("from")
            res = await self.client.search(index=index, **query_kwargs)

        elif isinstance(query, MultiSearch):
            await self.semaphore.acquire()
            try:
                res = await self.client.msearch(body=query.to_dict(), index=index)
            finally:
                self.semaphore.release()
            res = res["responses"]

        if options.get("raw"):
            raise RawResultInterrupt(res)

        return res


class MongoQueryBackend:
    def __init__(self, client, collections):
        self.client = client
        self.collections = collections

        if None not in self.collections:  # set default collection pattern
            self.collections[None] = next(iter(self.collections.values()))

    def execute(self, query, **options):
        client = self.client[self.collections[options.get("biothing_type")]]
        return list(client.find(*query).skip(options.get("from", 0)).limit(options.get("size", 10)))


class SQLQueryBackend:
    def __init__(self, client):
        self.client = client

    def execute(self, query, **options):
        result = self.client.execute(query)
        return result.keys(), result.all()



"""
    Search Result Formatter

    Transform the raw query result into consumption-friendly
    structures by possibly removing from, adding to, and/or
    flattening the raw response from the database engine for
    one or more individual queries.

"""

from collections import UserDict, defaultdict

from biothings.utils.common import dotdict, traverse
from biothings.utils.jmespath import options as jmp_options


class FormatterDict(UserDict):
    def collapse(self, key):
        self.update(self.pop(key))

    def exclude(self, keys):
        for key in keys:
            self.pop(key, None)

    def include(self, keys):
        for key in list(self.keys()):
            if key in keys:
                self.pop(key)

    def wrap(self, key, kls):
        if isinstance(self.get(key), list):
            self[key] = [kls(x) for x in self[key]]
        else:
            self[key] = kls(self[key])


class Hits(FormatterDict):
    """
    {
        "total": ... ,
        "hits": [
            { ... },
            { ... },
            ...
        ]
    }
    """


class Doc(FormatterDict):
    """
    {
        "_id": ... ,
        "_score": ... ,
        ...
    }
    """


class ResultFormatterException(Exception):
    pass


class ResultFormatter:
    def transform(self, response):
        return response

    def transform_mapping(self, mapping, prefix=None, search=None):
        return mapping


class ESResultFormatter(ResultFormatter):
    """Class to transform the results of the Elasticsearch query generated prior in the pipeline.
    This contains the functions to extract the final document from the elasticsearch query result
    in `Elasticsearch Query`_.  This also contains the code to flatten a document etc.
    """

    class _Hits(Hits):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # make sure the document is coming from
            # elasticsearch at initialization time
            assert "hits" in self.data
            assert "total" in self.data["hits"]
            assert "hits" in self.data["hits"]
            for hit in self.data["hits"]["hits"]:
                assert "_source" in hit

    class _Doc(Doc):
        pass

    def __init__(self, licenses=None, license_transform=None, field_notes=None, excluded_keys=()):
        # do not directly use the "or" syntax
        # to turn a None object to a {}, that
        # would also replace external empty {}
        # that will later get populated with
        # contents when data is available.

        if licenses is None:
            licenses = {}
        if license_transform is None:
            license_transform = {}
        if field_notes is None:
            field_notes = {}

        # license settings
        # ---------------------
        # this feature is turned on by default,
        # it appends license fields to the documents
        # by looking at its first level field names
        # or the transformed field name equivalences.

        self.licenses = licenses
        # example:
        # {
        #     "exac": "http://example.com/licenseA",
        #     "snpeff": "http://example.com/licenseB"
        # }
        self.license_transform = license_transform
        # example:
        # {
        #     "exac_nontcga": "exac",
        #     "snpeff.ann": "snpeff"
        # }

        # mapping dispaly settings
        # -------------------------
        self.field_notes = field_notes
        self.excluded_keys = excluded_keys

    # for compatibility
    traverse = staticmethod(traverse)

    def transform(self, response, **options):
        """
        Transform the query response to a user-friendly structure.
        Mainly deconstruct the elasticsearch response structure and
        hand over to transform_doc to apply the options below.

        Options:
            # generic transformations for dictionaries
            # ------------------------------------------
            dotfield: flatten a dictionary using dotfield notation
            _sorted: sort keys alaphabetically in ascending order
            always_list: ensure the fields specified are lists or wrapped in a list
            allow_null: ensure the fields specified are present in the result,
                        the fields may be provided as type None or [].

            # additional multisearch result transformations
            # ------------------------------------------------
            template: base dict for every result, for example: {"success": true}
            templates: a different base for every result, replaces the setting above
            template_hit: a dict to update every positive hit result, default: {"found": true}
            template_miss: a dict to update every query with no hit, default: {"found": false}

            # document format and content management
            # ---------------------------------------
            biothing_type: result document type to apply customized transformation.
                        for example, add license field basing on document type's metadata.
            one: return the individual document if there's only one hit. ignore this setting
                if there are multiple hits. return None if there is no hit. this option is
                not effective when aggregation results are also returned in the same query.
            native: bool, if the returned result is in python primitive types.
            version: bool, if _version field is kept.
            score: bool, if _score field is kept.
            with_total: bool, if True, the response will include max_total documents,
                and a message to tell how many query terms return greater than the max_size of hits.
                The default is False.
                An example when with_total is True:
                {
                    'max_total': 100,
                    'msg': '12 query terms return > 1000 hits, using from=1000 to retrieve the remaining hits',
                    'hits': [...]
                }
        """
        options = dotdict(options)
        if isinstance(response, list):
            max_total = 0
            count_query_exceed_max_size = 0
            # If options.set isn't set, the default number of hits returned by the ES is 10.
            # Ref: https://www.elastic.co/guide/en/elasticsearch/reference/current/search-multi-search.html#search-multi-search-api-request-body  # noqa: F501
            max_size = options.size or 10

            responses_ = []
            options.pop("one", None)  # ignore
            template = options.pop("template", {})
            templates = options.pop("templates", [template] * len(response))
            template_hit = options.pop("template_hit", dict(found=True))
            template_miss = options.pop("template_miss", dict(found=False))
            responses = [self.transform(res, **options) for res in response]
            for tpl, res in zip(templates, responses):
                if options.with_total:
                    total = res.get("total", {}).get("value") or 0
                    if total > max_total:
                        max_total = total
                    if total > max_size:
                        count_query_exceed_max_size += 1

                for _res in res if isinstance(res, list) else [res]:
                    assert isinstance(_res, dict)
                    if _res and "hits" not in _res:
                        hit_ = dict(tpl)
                        hit_.update(template_hit)
                        hit_.update(_res)
                        responses_.append(hit_)
                        continue
                    if not _res or not _res["hits"]:
                        tpl.update(template_miss)
                        responses_.append(tpl)
                        continue
                    for hit in _res["hits"]:
                        hit_ = dict(tpl)
                        hit_.update(template_hit)
                        hit_.update(hit)
                        responses_.append(hit_)
            response_ = list(filter(None, responses_))

            if options.with_total:
                response_ = {
                    "max_total": max_total,
                    "hits": response_,
                }
                if count_query_exceed_max_size > 0:
                    _from = (options.get("from") or 0) + max_size
                    response_["msg"] = (
                        f"{count_query_exceed_max_size} query terms return > {max_size} hits, "
                        f"using from={_from} to retrieve the remaining hits"
                    )

            return response_

        if isinstance(response, dict):
            response = self._Hits(response)
            response.collapse("hits")
            response.exclude(("_shards", "_node", "timed_out"))
            response.wrap("hits", self._Doc)

            for hit in response["hits"]:
                hit.collapse("_source")
                # 'sort' is introduced when sorting
                hit.exclude(("_index", "_type", "sort"))
                self._transform_hit(hit, options)

            if options.get("native", True):
                response["hits"] = [hit.data for hit in response["hits"]]
                response = response.data

            if "aggregations" in response:
                self.transform_aggs(response["aggregations"])
                response["facets"] = response.pop("aggregations")
                hits = response.pop("hits")  # move key order
                if hits:  # hide "hits" field when size=0
                    response["hits"] = hits

            elif options.get("one"):
                # prefer one-level presentation
                # or structures as simple as possible
                if len(response["hits"]) == 1:
                    response = response["hits"][0]
                elif len(response["hits"]) == 0:
                    response = None
                else:  # show a list of docs
                    response = response["hits"]

            return response

        raise TypeError()

    def _transform_hit(self, doc, options):
        """
        In-place apply a variety of transformations to a document like:
        {
            "_id": ... ,
            "_index": ... ,
            ...
        }
        """
        if not options.get("version", False):
            doc.pop("_version", None)
        if not options.get("score", True):
            doc.pop("_score", None)
        for path, obj in self.traverse(doc):
            self.transform_hit(path, obj, options)
            if options.allow_null:
                self._allow_null(path, obj, options.allow_null)
            if options.always_list:
                self._always_list(path, obj, options.always_list)
            if options._sorted:
                self._sorted(path, obj)
        if options.dotfield:
            self._dotfield(doc, options)

    @staticmethod
    def _allow_null(path, obj, fields):
        """
        The specified fields should be set to None if it does not exist.
        When flattened, the field could be converted to an empty list.
        """
        if isinstance(obj, (dict, Doc)):
            for field in fields or []:
                if field.startswith(path):
                    key = field[len(path) :].lstrip(".")  # noqa: E203
                    if "." not in key and key not in obj:
                        obj[key] = None

    @staticmethod
    def _always_list(path, obj, fields):
        """
        The specified fields, if exist, should be set to a list type.
        None converts to an emtpy list [] instead of [None].
        """
        if isinstance(obj, (dict, Doc)):
            for field in fields:
                if field.startswith(path):
                    key = field[len(path) :].lstrip(".")  # noqa: E203
                    if key in obj and not isinstance(obj[key], list):
                        obj[key] = [obj[key]] if obj[key] is not None else []

    @staticmethod
    def _sorted(_, obj):
        """
        Sort a container in-place.
        """
        try:
            if isinstance(obj, (dict, Doc)):
                sorted_items = sorted(obj.items())
                obj.clear()
                obj.update(sorted_items)
        except Exception:
            pass

    @classmethod
    def _dotfield(cls, dic, options):
        """
        Flatten a dictionary.
        See biothings.utils.common.traverse for examples.
        """
        hit_ = defaultdict(list)
        for path, value in cls.traverse(dic, leaf_node=True):
            hit_[path].append(value)
        for key, lst in hit_.items():
            if len(lst) == 1 and key not in (options.always_list or []):
                hit_[key] = lst[0]
            else:  # multi-element list
                hit_[key] = list(filter(None, lst))
                if options._sorted:
                    cls._sorted(key, hit_[key])
        dic.clear()
        dic.update(hit_)

    def transform_hit(self, path, doc, options):
        """
        Transform an individual search hit result.
        By default add licenses for the configured fields.

        If a source has a license url in its metadata,
        Add "_license" key to the corresponding fields.
        Support dot field representation field alias.

        If we have the following settings in web_config.py

        LICENSE_TRANSFORM = {
            "exac_nontcga": "exac",
            "snpeff.ann": "snpeff"
        },

        Then GET /v1/variant/chr6:g.38906659G>A should look like:
        {
            "exac": {
                "_license": "http://bit.ly/2H9c4hg",
                "af": 0.00002471},
            "exac_nontcga": {
                "_license": "http://bit.ly/2H9c4hg",         <---
                "af": 0.00001883}, ...
        }
        And GET /v1/variant/chr14:g.35731936G>C could look like:
        {
            "snpeff": {
                "_license": "http://bit.ly/2suyRKt",
                "ann": [{"_license": "http://bit.ly/2suyRKt", <---
                        "effect": "intron_variant",
                        "feature_id": "NM_014672.3", ...},
                        {"_license": "http://bit.ly/2suyRKt", <---
                        "effect": "intron_variant",
                        "feature_id": "NM_001256678.1", ...}, ...]
            }, ...
        }

        The arrow marked fields would not exist without the setting lines.
        """

        licenses = self.licenses.get(options.biothing_type, {})
        if path in self.license_transform:
            path = self.license_transform[path]
        if path in licenses and isinstance(doc, dict):
            doc["_license"] = licenses[path]

        if options.jmespath:
            self.trasform_jmespath(path, doc, options)

    def trasform_jmespath(self, path: str, doc, options) -> None:
        """Transform any target field in doc using jmespath query syntax.
        The jmespath query parameter value should have the pattern of "<target_list_fieldname>|<jmespath_query_expression>"
        <target_list_fieldname> can be any sub-field of the input doc using dot notation, e.g. "aaa.bbb".
          If empty or ".", it will be the root field.
        The flexible jmespath syntax allows to filter/transform any nested objects in the input doc on the fly.
        The output of the jmespath transformation will then be used to replace the original target field value.
        Examples:
             * filtering an array sub-field
                 jmespath=tags|[?name==`Metadata`]  # filter tags array by name field
                 jmespath=aaa.bbb|[?(sub_a==`val_a`||sub_a==`val_aa`)%26%26sub_b==`val_b`]    # use %26%26 for &&
        """
        # options.jmespath is already validated and processed as a tuple
        # see biothings.web.options.manager.Coverter.translate
        parent_path, target_field, jmes_query = options.jmespath

        if path == parent_path:
            # we handle jmespath transformation at its parent field level,
            # so that we can set a transformed value
            target_field_value = doc.get(target_field) if target_field else doc
            if target_field_value:
                # pass jmp_options to include our own custom jmespath functions
                transformed_field_value = jmes_query.search(target_field_value, options=jmp_options)
                if target_field:
                    doc[target_field] = transformed_field_value
                else:
                    # if the target field is the root field, we need to replace the whole doc
                    # note that we cannot use `doc = transformed_field_value` here, because
                    # it will break the reference to the original doc object
                    doc.clear()
                    doc.update(transformed_field_value)

    def transform_aggs(self, res):
        """
        Transform the aggregations field and make it more presentable.
        For example, these are the fields of a two level nested aggregations:

            aggregations.<term>.doc_count_error_upper_bound
            aggregations.<term>.sum_other_doc_count
            aggregations.<term>.buckets.key
            aggregations.<term>.buckets.key_as_string
            aggregations.<term>.buckets.doc_count
            aggregations.<term>.buckets.<nested_term>.* (recursive)

        After the transformation, we'll have:

            facets.<term>._type
            facets.<term>.total
            facets.<term>.missing
            facets.<term>.other
            facets.<term>.terms.count
            facets.<term>.terms.term
            facets.<term>.terms.<nested_term>.* (recursive)

        Note the first level key change doesn't happen here.
        """

        for facet in res:
            res[facet]["_type"] = "terms"  # a type of ES Bucket Aggs
            res[facet]["terms"] = res[facet].pop("buckets")
            res[facet]["other"] = res[facet].pop("sum_other_doc_count", 0)
            res[facet]["missing"] = res[facet].pop("doc_count_error_upper_bound", 0)

            count = 0

            for bucket in res[facet]["terms"]:
                bucket["count"] = bucket.pop("doc_count")
                bucket["term"] = bucket.pop("key")
                if "key_as_string" in bucket:
                    bucket["term"] = bucket.pop("key_as_string")
                count += bucket["count"]

                # nested aggs
                for agg_k in list(bucket.keys()):
                    if isinstance(bucket[agg_k], dict):
                        bucket.update(self.transform_aggs(dict({agg_k: bucket[agg_k]})))

            res[facet]["total"] = count

        return res

    def transform_mapping(self, mapping, prefix=None, search=None):
        """
        Transform Elasticsearch mapping definition to
        user-friendly field definitions metadata results.
        """
        assert isinstance(mapping, dict)
        assert isinstance(prefix, str) or prefix is None
        assert isinstance(search, str) or search is None

        result = {}
        items = list(mapping.items())
        items.reverse()

        while items:
            key, dic = items.pop()
            dic = dict(dic)
            dic.pop("dynamic", None)
            dic.pop("normalizer", None)

            if key in self.field_notes:
                result["notes"] = self.field_notes[key]

            if "copy_to" in dic:
                if "all" in dic["copy_to"]:
                    dic["searched_by_default"] = True
                del dic["copy_to"]

            if "index" not in dic:
                if "enabled" in dic:
                    dic["index"] = dic.pop("enabled")
                else:  # true by default
                    dic["index"] = True

            if "properties" in dic:
                dic["type"] = "object"
                subs = ((".".join((key, k)), v) for k, v in dic["properties"].items())
                items.extend(reversed(list(subs)))
                del dic["properties"]

            if all(
                (
                    not self.excluded_keys or key not in self.excluded_keys,
                    not prefix or key.startswith(prefix),
                    not search or search in key,
                )
            ):
                result[key] = dict(sorted(dic.items()))

        return result


class MongoResultFormatter(ResultFormatter):
    def transform(self, result, **options):
        return {"total": {"value": len(result), "relation": "gte"}, "hits": result}


class SQLResultFormatter(ResultFormatter):
    def transform(self, result, **options):
        header, content = result
        return {
            "total": {"value": len(content), "relation": "gte"},
            "hits": [dict(zip(header, row)) for row in content],
        }



# NOTE
# THIS MODULE IS FOR REFERENCE ONLY
# AND IS NO LONGER COMPATIBLE WITH THE LATEST CODEBASE

"""
    Utils for creating swagger spec from a config_web module.

"""
from collections import OrderedDict

import yaml


def setup_yaml():
    """https://stackoverflow.com/a/8661021"""
    represent_dict_order = lambda self, data: self.represent_mapping("tag:yaml.org,2002:map", data.items())
    yaml.add_representer(OrderedDict, represent_dict_order)


all_params = {}


def _get_common_parameters(settings, biothing_object, url):
    _all = _get_all_params(settings)
    _ret = OrderedDict([])
    # get all params except for q, scopes, ids -> these are added to the request body, and not the query string
    _common_params_list = [
        (k, v)
        for (k, v) in _all.items()
        if (k != "q") and (k != "scopes") and (k != "ids") and (k in getattr(settings, "KWARG_DESCRIPTIONS", {}))
    ]
    # add 'special' parameters not controlled by main biothings web pipeline
    _common_params_list.extend(
        [("callback", {"type": str, "default": None}), ("email", {"type": str, "default": None})]
    )
    _description_templates = getattr(settings, "KWARG_DESCRIPTIONS", {})
    for param, _description in _common_params_list:
        _description_template_object = _description_templates.get(param)
        _param_display = _description_template_object.get("name")
        _ret.update(
            OrderedDict(
                [
                    (
                        _param_display,
                        OrderedDict(
                            [
                                ("name", _param_display),
                                ("in", "query"),
                                (
                                    "description",
                                    _description_template_object.get("text_template", "").format(
                                        biothing_object=biothing_object,
                                        param_type=_string_types(_all.get(param, {}).get("type", "")),
                                        param_default_value=_param_default_values(
                                            _all.get(param, {}).get("default", "")
                                        ),
                                        param_max=_param_max(_all.get(param, {}).get("max", "")),
                                    ),
                                ),
                                ("schema", {"type": _schema_types(_all.get(param, {}).get("type", ""))}),
                            ]
                        ),
                    )
                ]
            )
        )
    return _ret


def _get_schemas(settings, biothing_object, url):
    return OrderedDict(
        [
            (
                "string_or_array",
                {
                    "oneOf": [
                        {"type": "array", "items": {"type": "string"}},
                        {"type": "string"},
                    ]
                },
            ),
            (
                "int64_or_array",
                {
                    "oneOf": [
                        {"type": "array", "items": {"type": "integer", "format": "int64"}},
                        {"type": "integer", "format": "int64"},
                    ]
                },
            ),
            (
                "QueryResult",
                {
                    "type": "object",
                    "properties": OrderedDict(
                        [
                            ("max_score", {"type": "number", "format": "float"}),
                            ("took", {"type": "integer"}),
                            ("total", {"type": "integer"}),
                            (
                                "hits",
                                {
                                    "type": "array",
                                    "items": {"$ref": "#/components/schemas/{}".format(biothing_object.title())},
                                },
                            ),
                        ]
                    ),
                },
            ),
            (
                "QueryPOSTResult",
                {
                    "type": "array",
                    "items": {
                        "allOf": [
                            {"$ref": "#/components/schemas/{}".format(biothing_object.title())},
                            {
                                "type": "object",
                                "properties": {
                                    "_score": {"type": "number", "format": "float"},
                                    "query": {"type": "string"},
                                },
                            },
                        ]
                    },
                },
            ),
            (
                "ErrorResult",
                {"type": "object", "properties": {"success": {"type": "boolean"}, "message": {"type": "string"}}},
            ),
            (
                "{}".format(biothing_object.title()),
                {"type": "object", "required": ["_id"], "properties": {"_id": {"type": "string"}}},
            ),
        ]
    )


def _schema_types(t):
    # if t == type(bool()):
    if issubclass(t, bool):
        return "boolean"
    # if t == type(int()):
    if issubclass(t, int):
        return "integer"
    return "string"


def _string_types(t):
    # if t == type(str()):
    if issubclass(t, str):
        return " Type: string."
    # elif t == type(int()):
    elif issubclass(t, int):
        return " Type: integer."
    # elif t == type([]):
    elif issubclass(t, list):
        return " Type: string (list)."
    # elif t == type(bool()):
    elif issubclass(t, bool):
        return " Type: boolean."
    return " Type: string."


def _param_default_values(d):
    if not d:
        return " Default: None."
    # if type(d) == type(int()) or type(d) == type(str()):
    if isinstance(d, (int, str)):
        return " Default: {}.".format(d)
    # elif type(d) == type(True) and d:
    elif isinstance(d, bool) and d:
        return " Default: true."
    # elif type(d) == type(False) and not d:
    elif isinstance(d, bool) and not d:
        return " Default: false."
    return " Default: None."


def _param_max(m):
    if not m:
        return ""
    return " Max: {}.".format(m)


def _query_get_params(settings):
    _ret = {}
    for d in [
        settings.QUERY_GET_CONTROL_KWARGS,
        settings.QUERY_GET_ES_KWARGS,
        settings.QUERY_GET_ESQB_KWARGS,
        settings.QUERY_GET_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            if k != "q" and k in getattr(settings, "KWARG_DESCRIPTIONS", {}):
                _ret.setdefault(k, v)

    _ret["callback"] = {"type": str, "default": None}
    _ret["email"] = {"type": str, "default": None}

    return _ret


def _metadata_params(settings):
    _ret = {}
    for d in [
        settings.METADATA_GET_CONTROL_KWARGS,
        settings.METADATA_GET_ES_KWARGS,
        settings.METADATA_GET_ESQB_KWARGS,
        settings.METADATA_GET_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            if k in getattr(settings, "KWARG_DESCRIPTIONS", {}):
                _ret.setdefault(k, v)

    _ret["callback"] = {"type": str, "default": None}

    return _ret


def _annotation_get_params(settings):
    _ret = {}
    for d in [
        settings.ANNOTATION_GET_CONTROL_KWARGS,
        settings.ANNOTATION_GET_ES_KWARGS,
        settings.ANNOTATION_GET_ESQB_KWARGS,
        settings.ANNOTATION_GET_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            if k != "q" and k in getattr(settings, "KWARG_DESCRIPTIONS", {}):
                _ret.setdefault(k, v)

    _ret["callback"] = {"type": str, "default": None}
    _ret["email"] = {"type": str, "default": None}

    return _ret


def _query_post_params(settings):
    _ret = {}
    for d in [
        settings.QUERY_POST_CONTROL_KWARGS,
        settings.QUERY_POST_ES_KWARGS,
        settings.QUERY_POST_ESQB_KWARGS,
        settings.QUERY_POST_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            if k != "q" and k != "scopes" and k in getattr(settings, "KWARG_DESCRIPTIONS", {}):
                _ret.setdefault(k, v)

    _ret["email"] = {"type": str, "default": None}

    return _ret


def _annotation_post_params(settings):
    _ret = {}
    for d in [
        settings.ANNOTATION_POST_CONTROL_KWARGS,
        settings.ANNOTATION_POST_ES_KWARGS,
        settings.ANNOTATION_POST_ESQB_KWARGS,
        settings.ANNOTATION_POST_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            if k != "ids" and k in getattr(settings, "KWARG_DESCRIPTIONS", {}):
                _ret.setdefault(k, v)

    _ret["email"] = {"type": str, "default": None}

    return _ret


def _get_all_params(settings):
    # returns all possible parameters for a biothing web settings object
    if all_params:
        return all_params

    for d in [
        settings.ANNOTATION_GET_CONTROL_KWARGS,
        settings.ANNOTATION_GET_ES_KWARGS,
        settings.ANNOTATION_GET_ESQB_KWARGS,
        settings.ANNOTATION_GET_TRANSFORM_KWARGS,
        settings.ANNOTATION_POST_CONTROL_KWARGS,
        settings.ANNOTATION_POST_ES_KWARGS,
        settings.ANNOTATION_POST_ESQB_KWARGS,
        settings.ANNOTATION_POST_TRANSFORM_KWARGS,
        settings.QUERY_GET_CONTROL_KWARGS,
        settings.QUERY_GET_ES_KWARGS,
        settings.QUERY_GET_ESQB_KWARGS,
        settings.QUERY_GET_TRANSFORM_KWARGS,
        settings.QUERY_POST_CONTROL_KWARGS,
        settings.QUERY_POST_ES_KWARGS,
        settings.QUERY_POST_ESQB_KWARGS,
        settings.QUERY_POST_TRANSFORM_KWARGS,
        settings.METADATA_GET_CONTROL_KWARGS,
        settings.METADATA_GET_ES_KWARGS,
        settings.METADATA_GET_ESQB_KWARGS,
        settings.METADATA_GET_TRANSFORM_KWARGS,
    ]:
        for k, v in d.items():
            all_params.setdefault(k, v)

    return all_params


def _query_get_description(settings, biothing_object, url):
    _descriptions = getattr(settings, "KWARG_DESCRIPTIONS", {})
    _parameters = [
        OrderedDict(
            [
                ("name", "q"),
                ("in", "query"),
                ("description", _descriptions.get("q", {}).get("text_template").format(doc_query_syntax_url="")),
                ("required", True),
                ("example", getattr(settings, "STATUS_CHECK", {}).get("id", "EXAMPLE")),
                ("schema", {"type": "string"}),
            ]
        )
    ]
    for k in _query_get_params(settings).keys():
        _parameters.append(
            OrderedDict(
                [
                    ("name", _descriptions.get(k).get("name")),
                    ("$ref", "#/components/parameters/{}".format(_descriptions.get(k).get("name"))),
                ]
            )
        )
    return OrderedDict(
        [
            ("tags", ["query"]),
            (
                "summary",
                "Make {biothing_object} queries and return matching {biothing_object} hits. Supports JSONP and CORS as well.".format(
                    biothing_object=biothing_object
                ),
            ),
            ("parameters", _parameters),
            (
                "responses",
                OrderedDict(
                    [
                        (
                            "200",
                            OrderedDict(
                                [
                                    ("description", 'A query response with the "hits" property'),
                                    (
                                        "content",
                                        {"application/json": {"schema": {"$ref": "#/components/schemas/QueryResult"}}},
                                    ),
                                ]
                            ),
                        ),
                        (
                            "400",
                            OrderedDict(
                                [
                                    ("description", "A response indicating an improperly formatted query"),
                                    (
                                        "content",
                                        {"application/json": {"schema": {"$ref": "#/components/schemas/ErrorResult"}}},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
            ),
        ]
    )


def _query_post_description(settings, biothing_object, url):
    _descriptions = getattr(settings, "KWARG_DESCRIPTIONS", {})
    _parameters = []
    for k in _query_post_params(settings).keys():
        _parameters.append(
            OrderedDict(
                [
                    ("name", _descriptions.get(k).get("name")),
                    ("$ref", "#/components/parameters/{}".format(_descriptions.get(k).get("name"))),
                ]
            )
        )
    return OrderedDict(
        [
            ("tags", ["query"]),
            (
                "summary",
                "Make {biothing_object} batch queries and return matching {biothing_object} hits".format(
                    biothing_object=biothing_object
                ),
            ),
            (
                "requestBody",
                {
                    "content": {
                        "application/x-www-form-urlencoded": {
                            "schema": {
                                "properties": OrderedDict(
                                    [
                                        (
                                            "q",
                                            OrderedDict(
                                                [
                                                    (
                                                        "description",
                                                        'multiple query terms separated by comma (also "+" or whitespace).  Does not support wildcard queries',
                                                    ),
                                                    ("type", "string"),
                                                ]
                                            ),
                                        ),
                                        (
                                            "scopes",
                                            OrderedDict(
                                                [
                                                    (
                                                        "description",
                                                        _descriptions.get("scopes", {})
                                                        .get("text_template", "")
                                                        .format(param_type="string", param_max=""),
                                                    ),
                                                    ("type", "string"),
                                                ]
                                            ),
                                        ),
                                    ]
                                )
                            }
                        }
                    }
                },
            ),
            ("parameters", _parameters),
            (
                "responses",
                OrderedDict(
                    [
                        (
                            "200",
                            OrderedDict(
                                [
                                    ("description", 'Query response objects with the "hits" property'),
                                    (
                                        "content",
                                        {
                                            "application/json": {
                                                "schema": {"$ref": "#/components/schemas/QueryPOSTResult"}
                                            }
                                        },
                                    ),
                                ]
                            ),
                        ),
                        (
                            "400",
                            OrderedDict(
                                [
                                    ("description", "A response indicating an improperly formatted query"),
                                    (
                                        "content",
                                        {"application/json": {"schema": {"$ref": "#/components/schemas/ErrorResult"}}},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
            ),
        ]
    )


def _annotation_get_description(settings, biothing_object, url):
    _descriptions = getattr(settings, "KWARG_DESCRIPTIONS", {})
    _parameters = [
        OrderedDict(
            [
                ("name", "{biothing_object}id".format(biothing_object=biothing_object)),
                ("in", "path"),
                ("description", "{biothing_object} id".format(biothing_object=biothing_object)),
                ("required", True),
                ("example", getattr(settings, "STATUS_CHECK", {}).get("id", "EXAMPLE")),
                ("schema", {"type": "string"}),
            ]
        )
    ]
    for k in _annotation_get_params(settings).keys():
        _parameters.append(
            OrderedDict(
                [
                    ("name", _descriptions.get(k).get("name")),
                    ("$ref", "#/components/parameters/{}".format(_descriptions.get(k).get("name"))),
                ]
            )
        )
    return OrderedDict(
        [
            ("tags", ["{}".format(biothing_object)]),
            (
                "summary",
                "Retrieve {biothing_object} objects based on {biothing_object} id.  Supports JSONP and CORS as well.".format(
                    biothing_object=biothing_object
                ),
            ),
            ("parameters", _parameters),
            (
                "responses",
                OrderedDict(
                    [
                        (
                            "200",
                            OrderedDict(
                                [
                                    (
                                        "description",
                                        "A matching {biothing_object} object".format(biothing_object=biothing_object),
                                    ),
                                    (
                                        "content",
                                        {
                                            "application/json": {
                                                "schema": {
                                                    "$ref": "#/components/schemas/{}".format(biothing_object.title())
                                                }
                                            }
                                        },
                                    ),
                                ]
                            ),
                        ),
                        (
                            "404",
                            OrderedDict(
                                [
                                    (
                                        "description",
                                        "A response indicating an unknown {biothing_object} id".format(
                                            biothing_object=biothing_object
                                        ),
                                    )
                                ]
                            ),
                        ),
                    ]
                ),
            ),
        ]
    )


def _annotation_post_description(settings, biothing_object, url):
    _descriptions = getattr(settings, "KWARG_DESCRIPTIONS", {})
    _parameters = []
    for k in _annotation_post_params(settings).keys():
        _parameters.append(
            OrderedDict(
                [
                    ("name", _descriptions.get(k).get("name")),
                    ("$ref", "#/components/parameters/{}".format(_descriptions.get(k).get("name"))),
                ]
            )
        )
    return OrderedDict(
        [
            ("tags", ["{biothing_object}".format(biothing_object=biothing_object)]),
            (
                "summary",
                "For a list of {biothing_object} ids, return the matching {biothing_object} object".format(
                    biothing_object=biothing_object
                ),
            ),
            (
                "requestBody",
                {
                    "content": {
                        "application/x-www-form-urlencoded": {
                            "schema": OrderedDict(
                                [
                                    (
                                        "properties",
                                        {
                                            "ids": OrderedDict(
                                                [
                                                    (
                                                        "description",
                                                        _descriptions.get("ids", {})
                                                        .get("text_template", "")
                                                        .format(
                                                            biothing_object=biothing_object,
                                                            param_type=" Type: string (list).",
                                                            param_default_value="",
                                                            param_max=_param_max(
                                                                getattr(
                                                                    settings, "ANNOTATION_POST_CONTROL_KWARGS", {}
                                                                ).get("max", 1000)
                                                            ),
                                                        ),
                                                    ),
                                                    ("type", "string"),
                                                ]
                                            )
                                        },
                                    ),
                                    ("required", ["ids"]),
                                ]
                            )
                        }
                    }
                },
            ),
            ("parameters", _parameters),
            (
                "responses",
                OrderedDict(
                    [
                        (
                            "200",
                            OrderedDict(
                                [
                                    (
                                        "description",
                                        "A list of matching {biothing_object} objects".format(
                                            biothing_object=biothing_object
                                        ),
                                    ),
                                    (
                                        "content",
                                        {
                                            "application/json": {
                                                "schema": {
                                                    "type": "array",
                                                    "items": {
                                                        "$ref": "#/components/schemas/{}".format(
                                                            biothing_object.title()
                                                        )
                                                    },
                                                }
                                            }
                                        },
                                    ),
                                ]
                            ),
                        ),
                        (
                            "400",
                            OrderedDict(
                                [
                                    ("description", "A response indicating an improperly formatted query"),
                                    (
                                        "content",
                                        {"application/json": {"schema": {"$ref": "#/components/schemas/ErrorResult"}}},
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
            ),
        ]
    )


def _metadata_description(settings, biothing_object, url):
    return OrderedDict(
        [
            ("tags", ["metadata"]),
            ("summary", "Get metadata about the data available from {url}".format(url=url)),
            (
                "parameters",
                [OrderedDict([("name", "callback"), ("in", "query"), ("$ref", "#/components/parameters/callback")])],
            ),
            ("responses", {"200": {"description": "{url} metadata object".format(url=url)}}),
        ]
    )


def _metadata_fields_description(settings, biothing_object, url):
    _descriptions = getattr(settings, "KWARG_DESCRIPTIONS", {})
    _parameters = []
    for k in _metadata_params(settings).keys():
        _parameters.append(
            OrderedDict(
                [
                    ("name", _descriptions.get(k).get("name")),
                    ("$ref", "#/components/parameters/{}".format(_descriptions.get(k).get("name"))),
                ]
            )
        )
    return OrderedDict(
        [
            ("tags", ["metadata"]),
            (
                "summary",
                "Get metadata about the data fields available from a {url} {biothing_object} object".format(
                    biothing_object=biothing_object, url=url
                ),
            ),
            ("parameters", _parameters),
            ("responses", {"200": {"description": "{url} metadata fields object".format(url=url)}}),
        ]
    )


def create_openapi_json(_settings, **kwargs):
    _url = getattr(_settings, "GA_TRACKER_URL", "MyBiothing.info")
    _biothing_object = getattr(_settings, "ES_DOC_TYPE", "biothing")
    _doc = OrderedDict(
        [
            ("openapi", kwargs.get("openapi", "3.0.0")),
            (
                "info",
                OrderedDict(
                    [
                        (
                            "version",
                            kwargs.get(
                                "info__version",
                                "{:.1f}".format(float(getattr(_settings, "API_VERSION", "v1").lstrip("v"))),
                            ),
                        ),
                        (
                            "title",
                            kwargs.get(
                                "info__title", "{} API".format(getattr(_settings, "GA_TRACKER_URL", "MyBiothing.info"))
                            ),
                        ),
                        (
                            "description",
                            kwargs.get(
                                "info__description",
                                getattr(_settings, "KWARG_DESCRIPTION", {}).get(
                                    "_root",
                                    "Documentation of the {url} {biothing_object} query web services.  Learn more about [{url}](http://{url}/)".format(
                                        url=_url, biothing_object=_biothing_object
                                    ),
                                ),
                            ),
                        ),
                        ("termsOfService", kwargs.get("info__termsOfService", "http://{url}/terms".format(url=_url))),
                        (
                            "contact",
                            OrderedDict(
                                [
                                    ("name", kwargs.get("info__contact__name", "Chunlei Wu")),
                                    ("x-role", kwargs.get("info__contact__x_role", "responsible developer")),
                                    ("email", kwargs.get("info__contact__email", "help@biothings.io")),
                                    ("x-id", kwargs.get("info__contact__x_id", "https://github.com/newgene")),
                                ]
                            ),
                        ),
                    ]
                ),
            ),
            (
                "servers",
                [
                    OrderedDict(
                        [
                            (
                                "url",
                                "http://{url}/{version}".format(
                                    url=_url, version=getattr(_settings, "API_VERSION", "v1")
                                ),
                            ),
                            ("description", "Production server"),
                        ]
                    ),
                    OrderedDict(
                        [
                            (
                                "url",
                                "https://{url}/{version}".format(
                                    url=_url, version=getattr(_settings, "API_VERSION", "v1")
                                ),
                            ),
                            ("description", "Encrypted Production server"),
                        ]
                    ),
                ],
            ),
            (
                "tags",
                [
                    {"name": "{biothing_object}".format(biothing_object=_biothing_object)},
                    {"name": "query"},
                    {"name": "metadata"},
                ],
            ),
            (
                "paths",
                OrderedDict(
                    [
                        (
                            "/{biothing_object}/{{{biothing_object}id}}".format(biothing_object=_biothing_object),
                            # annotation GET
                            OrderedDict(
                                [
                                    (
                                        "get",
                                        _annotation_get_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    )
                                ]
                            ),
                        ),
                        # annotation POST
                        (
                            "/{biothing_object}".format(biothing_object=_biothing_object),
                            OrderedDict(
                                [
                                    (
                                        "post",
                                        _annotation_post_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    )
                                ]
                            ),
                        ),
                        # query GET/POST
                        (
                            "/query",
                            OrderedDict(
                                [
                                    (
                                        "get",
                                        _query_get_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    ),
                                    (
                                        "post",
                                        _query_post_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    ),
                                ]
                            ),
                        ),
                        # metadata GET
                        (
                            "/metadata",
                            OrderedDict(
                                [
                                    (
                                        "get",
                                        _metadata_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    )
                                ]
                            ),
                        ),
                        # metadata/fields GET
                        (
                            "/metadata/fields",
                            OrderedDict(
                                [
                                    (
                                        "get",
                                        _metadata_fields_description(
                                            settings=_settings, biothing_object=_biothing_object, url=_url
                                        ),
                                    )
                                ]
                            ),
                        ),
                    ]
                ),
            ),
            (
                "components",
                OrderedDict(
                    [
                        (
                            "parameters",
                            _get_common_parameters(settings=_settings, biothing_object=_biothing_object, url=_url),
                        ),
                        ("schemas", _get_schemas(settings=_settings, biothing_object=_biothing_object, url=_url)),
                    ]
                ),
            ),
        ]
    )
    return _doc


def create_openapi_yaml(settings, yaml_file, **kwargs):
    """Create an openapi compliant spec for the biothing project whose settings are input.

    settings - a biothings web settings object
    yaml_file - the output file
    """
    setup_yaml()
    _json = create_openapi_json(settings, **kwargs)
    with open(yaml_file, "w") as outf:
        yaml.dump(_json, outf, default_flow_style=False)



"""
    Biothings API Launcher

    In this module, we have three framework-specific launchers
    and a command-line utility to provide both programmatic and
    command-line access to start Biothings APIs.

"""
import logging
import os
import sys
from pprint import pformat

import tornado.httpserver
import tornado.ioloop
import tornado.log
import tornado.web
from tornado.options import define, options

from biothings import __version__
from biothings.web.applications import BiothingsAPI
from biothings.web.settings import configs

logger = logging.getLogger(__name__)


class BiothingsAPIBaseLauncher:
    def __init__(self, config=None):
        # see biothings.web.settings.configs.load_module
        # for all supported ways to specify a config module
        logging.info("Biothings API %s", __version__)
        self.config = configs.load(config)  # for biothings APIs
        self.settings = dict(debug=False)  # for web frameworks

    def get_app(self):
        raise NotImplementedError()

    def get_server(self):
        raise NotImplementedError()

    def start(self, port=8000):
        raise NotImplementedError()


class TornadoAPILauncher(BiothingsAPIBaseLauncher):
    # tornado uses its own event loop which is
    # a wrapper around the asyncio event loop

    def __init__(self, config=None):
        # About debug mode in tornado:
        # https://www.tornadoweb.org/en/stable/guide/running.html \
        # #debug-mode-and-automatic-reloading
        super().__init__(config)
        self.handlers = []  # additional handlers
        self.host = None

    def _configure_logging(self):
        root_logger = logging.getLogger()

        if isinstance(self.config, configs.ConfigPackage):
            config = self.config.root
        else:  # configs.ConfigModule
            config = self.config

        if hasattr(config, "LOGGING_FORMAT"):
            for handler in root_logger.handlers:
                if isinstance(handler.formatter, tornado.log.LogFormatter):
                    handler.formatter._fmt = config.LOGGING_FORMAT

        logging.getLogger("urllib3").setLevel(logging.ERROR)
        logging.getLogger("elasticsearch").setLevel(logging.WARNING)

        if self.settings["debug"]:
            root_logger.setLevel(logging.DEBUG)
        else:
            root_logger.setLevel(logging.INFO)

    @staticmethod
    def use_curl():
        """
        Use curl implementation for tornado http clients.
        More on https://www.tornadoweb.org/en/stable/httpclient.html
        """
        tornado.httpclient.AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

    def get_app(self):
        return BiothingsAPI.get_app(self.config, self.settings, self.handlers)

    def get_server(self):
        # Use case example:
        # Run API in an external event loop.
        app = self.get_app()
        logger.info("All Handlers:\n%s", pformat(app.biothings.handlers, width=200))
        return tornado.httpserver.HTTPServer(app, xheaders=True)

    def start(self, port=8000):
        self._configure_logging()

        http_server = self.get_server()
        http_server.listen(port, self.host)

        logger.info('Server is running on "%s:%s"...', self.host or "0.0.0.0", port)
        loop = tornado.ioloop.IOLoop.instance()
        loop.start()


# WSGI
class FlaskAPILauncher(BiothingsAPIBaseLauncher):
    # Proof of concept
    # Not fully implemented

    # Create the following file under an application folder
    # to serve the application with a WSGI HTTP Server
    # like Gunicorn or use with AWS Elastic Beanstalk *

    # - application.py
    # from biothings.web.launcher import FlaskAPILauncher
    # application = FlaskAPILauncher("config").get_app()

    # * https://docs.aws.amazon.com/elasticbeanstalk/latest/dg/create-deploy-python-apps.html

    def get_app(self):
        from biothings.web.applications import FlaskBiothingsAPI

        return FlaskBiothingsAPI.get_app(self.config)

    def get_server(self):
        raise NotImplementedError()
        # https://flask.palletsprojects.com/en/2.0.x/deploying/wsgi-standalone/
        # from gevent.pywsgi import WSGIServer
        # return WSGIServer(('', 5000), self.get_app())

    def start(self, port=8000, dev=True):
        if dev:
            app = self.get_app()
            app.run(port=port)

        # example implementation
        # for gevent WSGI server
        else:
            server = self.get_server()
            server.serve_forever()


# ASGI
class FastAPILauncher(BiothingsAPIBaseLauncher):
    # Proof of concept
    # Not fully implemented

    # from biothings.web.launcher import FastAPILauncher
    # app = FastAPILauncher("config").get_app()

    # >>> uvicorn main:app --host 0.0.0.0 --port 80
    # INFO: Uvicorn running on http://0.0.0.0:80 (Press CTRL+C to quit)

    def get_app(self):
        from biothings.web.applications import FastAPIBiothingsAPI

        return FastAPIBiothingsAPI.get_app(self.config)


BiothingsAPILauncher = TornadoAPILauncher


# Command Line Utilities
# --------------------------

define("port", default=8000, help="run on the given port")
define("debug", default=False, help="debug settings like logging preferences")
define("address", default=None, help="host address to listen to, default to all interfaces")
define("autoreload", default=False, help="auto reload the web server when file change detected")
define("framework", default="tornado", help="the web freamework to start a web server")
define("conf", default="config", help="specify a config module name to import")
define("dir", default=os.getcwd(), help="path to app directory that includes config.py")


def main(app_handlers=None, app_settings=None, use_curl=False):
    """Start a Biothings API Server"""

    options.parse_command_line()
    _path = os.path.abspath(options.dir)
    if _path not in sys.path:
        sys.path.append(_path)
    del _path

    app_handlers = app_handlers or []
    app_settings = app_settings or {}

    if options.framework == "tornado":
        launcher = TornadoAPILauncher(options.conf)
    elif options.framework == "flask":
        launcher = FlaskAPILauncher(options.conf)
    elif options.framework == "fastapi":
        launcher = FastAPILauncher(options.conf)
    else:  # there are only three supported frameworks for now
        raise ValueError("Unsupported framework.")

    try:
        if app_settings:
            launcher.settings.update(app_settings)
        if app_handlers:
            launcher.handlers = app_handlers
        if use_curl:
            launcher.use_curl()

        launcher.host = options.address
        launcher.settings.update(debug=options.debug)
        launcher.settings.update(autoreload=options.autoreload)
    except Exception:
        pass

    launcher.start(options.port)


if __name__ == "__main__":
    main()



import datetime
import logging

# pandas.io.json encoder to deal with non-json compliant values NaN, Inf
# (based on ujson, but pandas has its own way to deal with these values)
# see https://github.com/biothings/biothings.api/commit/59c0d78f758018b0d87836657a2b5d1a700503a1
# import pandas.io.json as pdjson
# replace pandas json encoder with orjson:
import orjson
from tornado.web import RequestHandler

from biothings import config


class DefaultHandler(RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Content-Type", "application/json")
        # part of pre-flight requests
        self.set_header("Access-Control-Allow-Methods", "PUT, DELETE, POST, GET, OPTIONS")
        self.set_header("Access-Control-Allow-Headers", "Content-Type,X-BioThings-API,X-Biothings-Access-Token")

    def write(self, result):
        super(DefaultHandler, self).write(
            # pdjson.dumps({
            #     "result": result,
            #     "status": "ok"
            # }, iso_dates=True)
            orjson.dumps(
                {"result": result, "status": "ok"},
                option=orjson.OPT_NON_STR_KEYS | orjson.OPT_NAIVE_UTC,
            ).decode()
        )

    def write_error(self, status_code, **kwargs):
        self.set_status(status_code)
        super(DefaultHandler, self).write(
            {
                "error": str(kwargs.get("exc_info", [None, None, None])[1]),
                "status": "error",
                "code": status_code,
            }
        )

    # defined by default so we accept OPTIONS pre-flight requests
    def options(self, *args, **kwargs):
        logging.debug("OPTIONS args: %s, kwargs: %s" % (args, kwargs))


class BaseHandler(DefaultHandler):
    def initialize(self, managers, **kwargs):
        self.managers = managers


class GenericHandler(DefaultHandler):
    def initialize(self, shell, **kwargs):
        self.shell = shell

    def get(self, *args, **kwargs):
        logging.debug("GET args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method GET not allowed", None))

    def post(self, *args, **kwargs):
        logging.debug("POST args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method POST not allowed", None))

    def put(self, *args, **kwargs):
        logging.debug("PUT args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method PUT not allowed", None))

    def delete(self, *args, **kwargs):
        logging.debug("DELETE args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method DELETE not allowed", None))

    def head(self, *args, **kwargs):
        logging.debug("HEAD args: %s, kwargs: %s" % (args, kwargs))
        self.write_error(405, exc_info=(None, "Method HEAD not allowed", None))


class RootHandler(DefaultHandler):
    def initialize(self, features, hub_name=None, **kwargs):
        self.features = features
        self.hub_name = hub_name

    async def get(self):
        self.write(
            {
                "name": self.hub_name or getattr(config, "HUB_NAME", None),
                "biothings_version": getattr(config, "BIOTHINGS_VERSION", None),
                "app_version": getattr(config, "APP_VERSION", None),
                "icon": getattr(config, "HUB_ICON", None),
                "now": datetime.datetime.now().astimezone(),
                "features": self.features,
            }
        )



"""
    A build config contains autobuild configs and other information.

    TODO: not all features already supported in the code

    For exmaple:
    {
        "_id": "mynews",
        "name": "mynews",
        "doc_type": "news",
        "sources": ["mynews"],
        "root": ["mynews"],
        "builder_class": "biothings.hub.databuild.builder.DataBuilder",
        "autobuild": { ... },
        "autopublish": { ... },
        "build_version": "%Y%m%d%H%M"
    }

    Autobuild:
    - build
    - diff/snapshot

    Autopublish:
    - release note
    - publish

    Autorelease:
    - release

"""

try:
    from functools import singledispatchmethod
except ImportError:
    from singledispatchmethod import singledispatchmethod

import logging

logger = logging.getLogger(__name__)


class AutoBuildConfigError(Exception):
    pass


class AutoBuildConfig:
    """
    Parse automation configurations after each steps following 'build'.

    Example:
    {
        "autobuild": {
            "schedule": "0 8 * * 7", // Make a build every 08:00 on Sunday.
            "type": "diff",          // Auto create a "diff" w/previous version.
                                     // The other option is "snapshot".
            "env": "local",          // ES env to create an index and snapshot,
                                     // not required when type above is diff.
                                     // Setting the env also implies auto snapshot.
                                     // It could be in addition to auto diff.
                                     // Also accept (indexer_env, snapshot_env).
        },
        "autopublish": {
            "type": "snapshot",      // Auto publish new snapshots for new builds.
                                     // The type field can also be 'diff'.
            "env": "prod",           // The release environment to publish snapshot.
                                     // Or the release environment to publish diff.
                                     // This field is required for either type.
            "note": True             // If we should publish with a release note
                                     // TODO not implemented yet
        },
        "autorelease": {
            "schedule": "0 0 * * 1", // Make a release every Monday at midnight
                                     // (if there's a new published version.)
            "type": "full",          // Only auto install full releases.
                                     // The release type can also be 'incremental'.
        }
    }
    The terms below are used interchangeably.
    """

    BUILD_TYPES = ("diff", "snapshot")
    RELEASE_TYPES = ("incremental", "full")
    RELEASE_TO_BUILD = dict(zip(RELEASE_TYPES, BUILD_TYPES))

    def __init__(self, confdict):
        try:
            self.auto_build = confdict.get("autobuild", {})
            self.auto_build["type"] = self._types_as_set(self.auto_build.get("type"))

            self.auto_publish = confdict.get("autopublish", {})
            self.auto_publish["type"] = self._types_as_set(self.auto_publish.get("type"))

            self.auto_release = confdict.get("autorelease", {})
            self.auto_release["type"] = self._types_as_set(self.auto_release.get("type"))

        except AttributeError as exc:
            raise AutoBuildConfigError("Autobuild config entries should be dicts.") from exc

    @classmethod
    def _standardize_type(cls, type_str):
        """
        Ensure using build type 'diff' and 'snapshot'.
        Translate release type names accordingly.
        """
        if type_str in cls.BUILD_TYPES:
            return type_str
        elif type_str in cls.RELEASE_TYPES:
            return cls.RELEASE_TO_BUILD[type_str]
        raise AutoBuildConfigError(f"Unsupported type: {type_str}")

    @singledispatchmethod
    @classmethod
    def _types_as_set(cls, type_val):
        """
        The field "type", if specified, can be:
        "diff" or "incremental", "snapshot" or "full",
        or a list of the values mentioned above.
        """
        raise AutoBuildConfigError("Auto build type must be a list or string.")

    @_types_as_set.register(type(None))
    @classmethod
    def _(cls, _):
        return set()

    @_types_as_set.register(str)
    @classmethod
    def _(cls, type_str):
        return set((cls._standardize_type(type_str),))

    @_types_as_set.register(list)
    @classmethod
    def _(cls, type_list):
        return set(cls._standardize_type(_type) for _type in type_list)

    def export(self):
        autoconf = {
            "autobuild": dict(self.auto_build),
            "autopublish": dict(self.auto_publish),
            "autorelease": dict(self.auto_release),
        }
        autoconf["autobuild"]["type"] = list(autoconf["autobuild"]["type"])
        autoconf["autopublish"]["type"] = list(autoconf["autopublish"]["type"])
        autoconf["autorelease"]["type"] = list(autoconf["autorelease"]["type"])
        return autoconf

    # after build

    def should_diff_new_build(self):
        if "diff" in self.auto_build["type"]:
            return True

        if not self.auto_build["type"] and self.auto_build.get("schedule"):
            return True  # implicit

        # other implied relationships can be specified
        # --------------------------------------------
        # if 'diff' in self.auto_publish['type']:
        #     return True
        # if 'diff' in self.auto_release['type']:
        #     return True
        # --------------------------------------------
        # currently not sure if this is a good idea

        return False

    def should_snapshot_new_build(self):
        if "snapshot" in self.auto_build["type"] or self.auto_build.get("env"):
            # env indicates snapshot, diff doesn't need it
            if not self.auto_build.get("env"):
                logger.error("Auto snapshot env not set.")
            return True

        return False

    # after diff/snapshot

    def should_publish_new_diff(self):
        return "diff" in self.auto_publish["type"]

    def should_publish_new_snapshot(self):
        return "snapshot" in self.auto_publish["type"]

    # after publish

    def should_install_new_diff(self):
        return "diff" in self.auto_release["type"]

    def should_install_new_snapshot(self):
        return "snapshot" in self.auto_release["type"]

    def should_install_new_release(self):
        """
        Install the latest version regardless of update type/path.
        """
        if len(self.auto_release["type"]) == 2:
            return True

        if not self.auto_release["type"] and self.auto_release.get("schedule"):
            return True

        return False


def test():
    mygene = AutoBuildConfig({"autobuild": {"schedule": "0 8 * * 7"}})
    print(mygene.export())
    assert mygene.should_diff_new_build()
    assert not mygene.should_snapshot_new_build()
    assert not mygene.should_publish_new_diff()
    assert not mygene.should_publish_new_snapshot()
    assert not mygene.should_install_new_diff()
    assert not mygene.should_install_new_snapshot()
    assert not mygene.should_install_new_release()

    outbreak = AutoBuildConfig(
        {
            "autobuild": {
                "schedule": "0 8 * * *",
                "type": "snapshot",
                "env": "auto",  # snapshot location, indexing ES.
            },
            "autopublish": {"type": "snapshot", "env": "auto"},
            "autorelease": {"schedule": "0 10 * * *"},
        }
    )
    print(outbreak.export())
    assert not outbreak.should_diff_new_build()
    assert outbreak.should_snapshot_new_build()
    assert not outbreak.should_publish_new_diff()
    assert outbreak.should_publish_new_snapshot()
    assert not outbreak.should_install_new_diff()
    assert not outbreak.should_install_new_snapshot()
    assert outbreak.should_install_new_release()


if __name__ == "__main__":
    test()



import importlib
import importlib.util
import logging
import os
import pathlib
import sys

try:
    import typer
    from rich.logging import RichHandler

    typer_avail = True
except ImportError:
    typer_avail = False

from biothings.utils.common import DummyConfig
from biothings.utils.configuration import ConfigurationError

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

if typer_avail:
    # prevent dimming the help text from the 2nd line
    # see: https://github.com/tiangolo/typer/issues/437#issuecomment-1224149402
    typer.rich_utils.STYLE_HELPTEXT = ""

    # Typer already support an env variable to disable rich tracebacks using _TYPER_STANDARD_TRACEBACK=1
    # This supports a similar BTCLI_RICH_TRACEBACK=1 env variable to turn it on when default is off in our case
    # Relevant ref: https://github.com/tiangolo/typer/issues/525 and https://github.com/tiangolo/typer/discussions/612
    # and BTCLI_DEBUG=1 env varible to turn on both rich tracebacks and show locals
    if os.environ.get("BTCLI_DEBUG"):
        pretty_exceptions_enable = True
        pretty_exceptions_show_locals = True
        default_logging_level = logging.DEBUG
    else:
        if os.environ.get("BTCLI_RICH_TRACEBACK"):
            pretty_exceptions_enable = True
        else:
            pretty_exceptions_enable = False
            sys.tracebacklimit = 1  # only show the last traceback, default is 1000
        pretty_exceptions_show_locals = False
        default_logging_level = logging.INFO

    cli = typer.Typer(
        help="[green]BioThings Admin CLI to test your local data plugins. See helps for each command for specific usage.[/green]",
        rich_help_panel="Help and Others",
        rich_markup_mode="rich",
        context_settings=CONTEXT_SETTINGS,
        no_args_is_help=True,
        pretty_exceptions_show_locals=pretty_exceptions_show_locals,
        pretty_exceptions_enable=pretty_exceptions_enable,
    )

    logging.basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            # we don't need to turn on rich_tracebacks since typer creates it already
            RichHandler(
                level=default_logging_level,
                rich_tracebacks=False,
                tracebacks_suppress=[typer],
                show_path=False,
            ),
        ],
    )

logger = logging.getLogger("cli")


def setup_config():
    """Setup a config module necessary to launch the CLI"""
    working_dir = pathlib.Path().resolve()
    _config = DummyConfig("config")
    _config.HUB_DB_BACKEND = {
        "module": "biothings.utils.sqlite3",
        "sqlite_db_folder": ".biothings_hub",
    }
    _config.DATA_SRC_DATABASE = ".data_src_database"
    _config.DATA_ARCHIVE_ROOT = ".biothings_hub/archive"
    # _config.LOG_FOLDER = ".biothings_hub/logs"
    _config.LOG_FOLDER = None  # disable file logging, only log to stdout
    _config.DATA_PLUGIN_FOLDER = f"{working_dir}"
    _config.hub_db = importlib.import_module(_config.HUB_DB_BACKEND["module"])
    try:
        config_mod = importlib.import_module("config")
        for attr in dir(config_mod):
            value = getattr(config_mod, attr)
            if isinstance(value, ConfigurationError):
                raise ConfigurationError("%s: %s" % (attr, str(value)))
            setattr(_config, attr, value)
    except ModuleNotFoundError:
        logger.debug("The config.py does not exists in the working directory, use default biothings.config")

    sys.modules["config"] = _config
    sys.modules["biothings.config"] = _config


def main():
    """The main entry point for running the BioThings CLI to test your local data plugins."""
    if not typer_avail:
        logger.error(
            '"typer" package is required for CLI feature. Use "pip install biothings[cli]" or "pip install typer[all]" to install.'
        )
        return

    setup_config()
    from .dataplugin import app as dataplugin_app
    from .dataplugin_hub import app as dataplugin_hub_app

    cli.add_typer(dataplugin_app, name="dataplugin")
    cli.add_typer(dataplugin_hub_app, name="dataplugin-hub")
    return cli()



import json
import logging
import os
import sqlite3
from functools import wraps

from biothings.utils.common import find_value_in_doc, json_serial
from biothings.utils.dataload import update_dict_recur
from biothings.utils.dotfield import parse_dot_fields
from biothings.utils.hub_db import IDatabase
from biothings.utils.serializer import json_loads

config = None

logger = logging.getLogger(__name__)


def requires_config(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        global config
        if not config:
            try:
                from biothings import config as config_mod

                config = config_mod
            except ImportError:
                raise Exception("call biothings.config_for_app() first")
        return func(*args, **kwargs)

    return func_wrapper


@requires_config
def get_hub_db_conn():
    conn = DatabaseClient()
    return conn


@requires_config
def get_src_conn():
    conn = get_hub_db_conn()
    return conn


@requires_config
def get_src_db():
    conn = get_src_conn()
    return conn[config.DATA_SRC_DATABASE]


@requires_config
def get_src_dump():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "DATA_SRC_DUMP_COLLECTION", "src_dump")]


@requires_config
def get_src_master():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_MASTER_COLLECTION]


def get_src_build():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION]


def get_src_build_config():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_SRC_BUILD_COLLECTION + "_config"]


def get_data_plugin():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.DATA_PLUGIN_COLLECTION]


def get_api():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.API_COLLECTION]


def get_cmd():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][config.CMD_COLLECTION]


def get_event():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "EVENT_COLLECTION", "event")]


def get_hub_config():
    conn = get_hub_db_conn()
    return conn[config.DATA_HUB_DB_DATABASE][getattr(config, "HUB_CONFIG_COLLECTION", "hub_config")]


def get_last_command():
    try:
        db = get_cmd()
        res = db.get_conn().execute("SELECT MAX(_id) FROM cmd").fetchall()
        assert res[0][0], "No command ID found, bootstrap ?"
        return {"_id": res[0][0]}
    except Exception:
        return {"_id": 1}


def get_source_fullname(col_name):
    """
    Assuming col_name is a collection created from an upload process,
    find the main source & sub_source associated.
    """
    src_dump = get_src_dump()
    info = None
    for doc in src_dump.find():
        if col_name in doc.get("upload", {}).get("jobs", {}).keys():
            info = doc
    if info:
        name = info["_id"]
        if name != col_name:
            # col_name was a sub-source name
            return "%s.%s" % (name, col_name)
        else:
            return name


class Database(IDatabase):
    def __init__(self, db_folder, name=None):
        super(Database, self).__init__()
        if not name:
            self.name = config.DATA_HUB_DB_DATABASE
        else:
            self.name = name
        self.dbfile = os.path.join(db_folder, self.name)
        self.cols = {}

    @property
    def address(self):
        return self.dbfile

    def get_conn(self):
        return sqlite3.connect(self.dbfile)

    def collection_names(self):
        tables = self.get_conn().execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()
        return [name[0] for name in tables]

    def create_collection(self, colname):
        return self[colname]

    def create_if_needed(self, table):
        existings = [
            tname[0]
            for tname in self.get_conn()
            .execute("SELECT name FROM sqlite_master WHERE type='table' and " + "name = ?", (table,))
            .fetchall()
        ]
        if table not in existings:
            # TODO: injection...
            self.get_conn().execute("CREATE TABLE %s (_id TEXT PRIMARY KEY, document TEXT)" % table).fetchone()

    def __getitem__(self, colname):
        if colname not in self.cols:
            self.create_if_needed(colname)
            self.cols[colname] = Collection(colname, self)
        return self.cols[colname]


class DatabaseClient(IDatabase):
    def __init__(self):
        super().__init__()
        self.sqlite_db_folder = config.HUB_DB_BACKEND["sqlite_db_folder"]

        if not os.path.exists(self.sqlite_db_folder):
            os.makedirs(self.sqlite_db_folder)
        self.name = None
        self.dbfile = None
        self.cols = {}

    def __getitem__(self, name):
        return Database(self.sqlite_db_folder, name)


class Collection(object):
    def __init__(self, colname, db):
        self.colname = colname
        self.db = db

    def get_conn(self):
        return sqlite3.connect(self.db.dbfile)

    @property
    def name(self):
        return self.colname

    @property
    def database(self):
        return self.db

    def find_one(self, *args, **kwargs):
        if args and len(args) == 1 and isinstance(args[0], dict):
            if len(args[0]) == 1 and "_id" in args[0]:
                strdoc = (
                    self.get_conn()
                    .execute("SELECT document FROM %s WHERE _id = ?" % self.colname, (args[0]["_id"],))
                    .fetchone()
                )
                if strdoc:
                    return json_loads(strdoc[0])
                else:
                    return None
            else:
                return self.find(*args, find_one=True)
        elif args or kwargs:
            raise NotImplementedError("find(): %s %s" % (repr(args), repr(kwargs)))
        else:
            return self.find(find_one=True)

    def findv2(self, *args, **kwargs):
        """This is a new version of find() that uses json feature of sqlite3, will replace find in the future"""
        start = kwargs.get("start", 0)
        limit = kwargs.get("limit", 10)
        return_total = kwargs.get("return_total", False)  # return (results, total) tuple if True, default False
        return_list = kwargs.get("return_list", False)  # return list instead of generator if True, default False
        conn = self.get_conn()
        tbl_name = self.colname

        if args and len(args) == 1 and isinstance(args[0], dict) and len(args[0]) > 0:
            # it's key/value search, args[0] like {"a.b": "test", "a.b.c", "value"}
            sub_queries = []
            for k, v in args[0].items():
                if "*" in v or "?" in v:
                    _v = v.replace("*", "%").replace("?", "_")
                    _v = f"LIKE '{_v}'"
                else:
                    _v = f"= '{v}'"
                if k == "_id":
                    where = f"(_id {_v})"
                    sub_query = f"SELECT _id FROM {tbl_name} WHERE {where}"
                elif "." in k:
                    # nested field name like a.b.c, we will use json_tree.fullkey to match
                    # Here is an example for the query {"object.symbol": "BRD1"}:
                    # SELECT document FROM TISSUES, json_tree(TISSUES.document)
                    # WHERE (json_tree.fullkey LIKE '$.%object%.%symbol%' AND json_tree.value = 'BRD1')
                    k = k.replace(".", "%.%")
                    k = f"$.%{k}%"
                    where = f"(json_tree.fullkey LIKE '{k}' AND json_tree.value {_v})"
                    sub_query = f"SELECT _id FROM {tbl_name}, json_tree({tbl_name}.document) WHERE {where}"
                else:
                    # just a top level field, we will use json_each.key to match
                    # _v matches the value directly using LIKE or = (for a scalar field);
                    # _v2 matches the double-quoted value using LIKE (for an array field)
                    # Here is an example for the query {"ancestors": "CHEBI:75771"}:
                    # "SELECT _id FROM chebi, json_each(chebi.document)
                    # WHERE (json_each.key = 'ancestors' AND
                    #       (json_each.value = 'CHEBI:75771' OR json_each.value LIKE '%"CHEBI:75771"%')
                    # )
                    _v2 = _v.replace("LIKE '", "LIKE '%\"").replace("= '", "LIKE '%\"")
                    _v2 = _v2[:-1] + "\"%'"
                    where = f"(json_each.key = '{k}' AND (json_each.value {_v} OR json_each.value {_v2}))"
                    sub_query = f"SELECT _id FROM {tbl_name}, json_each({tbl_name}.document) WHERE {where}"
                sub_queries.append(sub_query)
            if sub_queries:
                if len(sub_queries) == 1:
                    query = sub_queries[0].replace("SELECT _id FROM", "SELECT document FROM")
                else:
                    # JOIN multiple sub queries:
                    # Here is an example for the query: q=object.symbol:BRD1%20AND%20subject.id:BTO:0000017
                    # SELECT document FROM TISSUES WHERE _id IN
                    #   (SELECT _id FROM
                    #       (SELECT _id FROM TISSUES, json_tree(TISSUES.document)
                    #           WHERE (json_tree.fullkey LIKE '$.%object%.%symbol%' AND json_tree.value = 'BRD1')
                    #       ) AS subq0
                    #       INNER JOIN
                    #       (SELECT _id FROM TISSUES, json_tree(TISSUES.document)
                    #           WHERE (json_tree.fullkey LIKE '$.%subject%.%id%' AND json_tree.value = 'BTO:0000017')
                    #       ) AS subq1
                    #       USING (_id)
                    #   )
                    query = f"SELECT _id FROM ({sub_queries[0]}) AS subq0"
                    for i, sub_query in enumerate(sub_queries[1:]):
                        query += f" INNER JOIN ({sub_query}) AS subq{i+1} USING (_id)"
                    query = f"SELECT document FROM {tbl_name} WHERE _id IN ({query})"
        elif not args or len(args) == 1 and len(args[0]) == 0:
            # nothing or empty dict
            query = f"SELECT document FROM {tbl_name}"
        else:
            raise NotImplementedError("find: args=%s kwargs=%s" % (repr(args), repr(kwargs)))

        # include limit and offset
        _query = query + f" LIMIT {limit} OFFSET {start}"
        logger.debug('SQLite query: "%s"', _query)
        results = (json_loads(doc[0]) for doc in conn.execute(_query))  # results is a generator
        if return_list:
            results = list(results)
        if return_total:
            # get total count without limit and offset
            total = conn.execute(query.replace("SELECT document FROM", "SELECT COUNT(*) FROM")).fetchone()[0]
            return results, total
        else:
            return results

    def find(self, *args, **kwargs):
        results = []
        if args and len(args) == 1 and isinstance(args[0], dict) and len(args[0]) > 0:
            # it's key/value search, let's iterate
            for doc in self.get_conn().execute("SELECT document FROM %s" % self.colname).fetchall():
                found = []
                doc = json_loads(doc[0])
                for k, v in args[0].items():
                    _found = find_value_in_doc(k, v, doc)
                    found.append(_found)
                if all(found):
                    if "find_one" in kwargs:
                        return doc
                    else:
                        results.append(doc)
            if "limit" in kwargs:
                start = kwargs.get("start", 0)
                end = start + kwargs.get("limit", 0)
                return results[start:end]
            return results
        elif not args or len(args) == 1 and len(args[0]) == 0:
            # nothing or empty dict
            results = [
                json_loads(doc[0])
                for doc in self.get_conn().execute("SELECT document FROM %s" % self.colname).fetchall()
            ]
            if "limit" in kwargs:
                start = kwargs.get("start", 0)
                end = start + kwargs.get("limit", 0)
                return results[start:end]
            return results
        else:
            raise NotImplementedError("find: args=%s kwargs=%s" % (repr(args), repr(kwargs)))

    def insert_one(self, doc):
        assert "_id" in doc
        with self.get_conn() as conn:
            conn.execute(
                "INSERT INTO %s (_id,document) VALUES (?,?)" % self.colname,
                (doc["_id"], json.dumps(doc, default=json_serial)),
            ).fetchone()
            conn.commit()

    def insert(self, docs, *args, **kwargs):
        with self.get_conn() as conn:
            for doc in docs:
                conn.execute(
                    "INSERT INTO %s (_id,document) VALUES (?,?)" % self.colname,
                    (doc["_id"], json.dumps(doc, default=json_serial)),
                ).fetchone()
                conn.commit()

    def bulk_write(self, docs, *args, **kwargs):
        doc_objs = [item._doc for item in docs]
        self.insert(doc_objs, *args, **kwargs)
        return Cursor(len(doc_objs))

    def update_one(self, query, what, upsert=False):
        assert len(what) == 1 and (
            "$set" in what or "$unset" in what or "$push" in what
        ), "$set/$unset/$push operators not found"
        doc = self.find_one(query)
        if doc:
            if "$set" in what:
                # parse_dot_fields uses json.dumps internally, we can to make
                # sure everything is serializable first
                what = json.loads(json.dumps(what, default=json_serial))
                what = parse_dot_fields(what["$set"])
                doc = update_dict_recur(doc, what)
            elif "$unset" in what:
                for keytounset in what["$unset"].keys():
                    doc.pop(keytounset, None)
            elif "$push" in what:
                for listkey, elem in what["$push"].items():
                    assert "." not in listkey, "$push not supported for nested keys: %s" % listkey
                    doc.setdefault(listkey, []).append(elem)
            self.save(doc)
        elif upsert:
            assert "$set" in what
            query.update(what["$set"])
            self.save(query)

    def update(self, query, what, upsert=False):
        docs = self.find(query)
        for doc in docs:
            self.update_one({"_id": doc["_id"]}, what, upsert)

    def save(self, doc):
        if self.find_one({"_id": doc["_id"]}):
            with self.get_conn() as conn:
                conn.execute(
                    "UPDATE %s SET document = ? WHERE _id = ?" % self.colname,
                    (json.dumps(doc, default=json_serial), doc["_id"]),
                )
                conn.commit()
        else:
            self.insert_one(doc)

    def replace_one(self, query, doc, upsert=False):
        assert "_id" in query
        orig = self.find_one(query)
        if orig:
            orig["_id"] = query["_id"]
            with self.get_conn() as conn:
                conn.execute(
                    "UPDATE %s SET document = ? WHERE _id = ?" % self.colname,
                    (json.dumps(doc, default=json_serial), orig["_id"]),
                )
                conn.commit()
        elif upsert:
            doc["_id"] = query["_id"]
            self.save(doc)

    def remove(self, query):
        docs = self.find(query)
        with self.get_conn() as conn:
            for doc in docs:
                conn.execute("DELETE FROM %s WHERE _id = ?" % self.colname, (doc["_id"],)).fetchone()
            conn.commit()

    def rename(self, new_name, dropTarget=False):
        with self.get_conn() as conn:
            if dropTarget:
                conn.execute(f"DROP TABLE IF EXISTS {new_name}")
            conn.execute(f"ALTER TABLE {self.colname} RENAME TO {new_name}").fetchall()

    def count(self):
        return self.get_conn().execute("SELECT count(_id) FROM %s" % self.colname).fetchone()[0]

    def drop(self):
        self.get_conn().execute("DROP TABLE %s" % self.colname).fetchall()

    def __getitem__(self, _id):
        return self.find_one({"_id": _id})

    def __getstate__(self):
        self.__dict__.pop("db", None)
        return self.__dict__


class Cursor(object):
    def __init__(self, inserted_count):
        self.inserted_count = inserted_count



""" Functions to return versions of things. """
import functools
import logging

# import pip
import os
import re
import shlex
import sys
from subprocess import DEVNULL, check_output

from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo

import biothings
from biothings.utils.dataload import dict_sweep


def get_python_version():
    """Get a list of python packages installed and their versions."""
    try:
        output = check_output([sys.executable or "python3", "-m", "pip", "list"], stderr=DEVNULL)
        return output.decode("utf-8").replace("\r", "").split("\n")[2:-1]
    except Exception:
        return []


@functools.lru_cache()
def get_biothings_commit():
    """Gets the biothings commit information."""
    try:
        with open(os.path.join(os.path.dirname(biothings.__file__), ".git-info"), "r", encoding="utf-8") as git_file:
            lines = [ln.strip("\n") for ln in git_file.readlines()]
            return {
                "repository-url": lines[0],
                "commit-hash": lines[1],
                "master-commits": lines[2],
                "version": biothings.__version__,
            }
    except Exception:
        return {
            "repository-url": "",
            "commit-hash": "",
            "master-commits": "",
            "version": biothings.__version__,
        }


@functools.lru_cache()
def get_repository_information(app_dir=None):
    """
    Get the repository information for the local repository, if it exists.
    """
    commit_hash = ""
    repository_url = ""

    if app_dir:
        app_dir = os.path.abspath(app_dir)

        try:
            args = shlex.split("git rev-parse HEAD")
            output = check_output(args, cwd=app_dir, stderr=DEVNULL)
            commit_hash = output.decode("utf-8").strip("\n")
        except Exception:
            pass

        try:
            args = shlex.split("git config --get remote.origin.url")
            output = check_output(args, cwd=app_dir, stderr=DEVNULL)
            repository_url = output.decode("utf-8").strip("\n")
        except Exception:
            pass

    codebase = {
        "repository-url": repository_url,
        "commit-hash": commit_hash,
    }
    return codebase


def get_python_exec_version():
    """return Python version"""
    return {
        "version": sys.version,
        "version_info": {
            "major": sys.version_info[0],
            "minor": sys.version_info[1],
            "micro": sys.version_info[2],
        },
    }


@functools.lru_cache()
def get_software_info(app_dir=None):
    """return current application info"""
    return {
        "python-package-info": get_python_version(),
        "codebase": get_repository_information(app_dir=app_dir),
        "biothings": get_biothings_commit(),
        "python-info": get_python_exec_version(),
    }


def check_new_version(folder, max_commits=10):
    """
    Given a folder pointing to a Git repo, return a dict containing info
    about remote commits not qpplied yet to the repo, or empty dict if nothing
    new.
    """
    # from https://stackoverflow.com/questions/8290233/gitpython-get-list-of-remote-commits-not-yet-applied
    try:
        repo = Repo(folder)
    except InvalidGitRepositoryError:
        logging.warning("Not a valid git repository for folder '%s', skipped for checking new version.", folder)
        return

    try:
        # Get URL from actual remote branch name that is being tracked.
        # more details: see comments in get_version
        remote_name = repo.active_branch.tracking_branch().remote_name
        url = repo.remote(remote_name).url
        repo_url = re.sub(r"\.git$", "", url)
    except Exception as err:
        logging.debug("Can't determine repository URL: %s", err)
        repo_url = None
    new_info = {}
    try:
        # we can't directly get the list of new commits without fetching them locally first
        # but we'd like to avoid fetching all the time just to check.
        # what we can do is a ls-remote and check the HEAD hash, if different, then fetch
        # (no pull) and inspect differences.
        try:
            head = repo.head.ref
        except TypeError as err:
            # cannot get the head reference, e.g. when HEAD is detached at certain commit point
            # For example, TypeError: HEAD is a detached symbolic reference as it points to '19ad50463d0cdd3329618789040c4b9012ccca24'
            logging.warning("%s, skipped for checking new version.", err)
            return

        tracking = head.tracking_branch()
        # inspect remote HEAD for that branch
        output = repo.git.ls_remote("--heads", tracking.remote_name, tracking.remote_head)
        remote_head_hexsha = output.split("\t")[0]
        if remote_head_hexsha == head.commit.hexsha:
            # hashes the same, we're up-to-date with the remote
            return
        else:
            logging.info("HEAD on remote is different, new commit(s) available for '%s'", folder)
            logging.info("HEAD(remote): %s, HEAD(local): %s", remote_head_hexsha, head.commit.hexsha)
            # need to fetch new code locally
            # usually one remotes, but just in case...
            for remote in repo.remotes:
                remote.fetch()
        # now identify new commits
        new_commits = [commit for commit in tracking.commit.iter_items(repo, f"{head.path}..{tracking.path}")]
        if new_commits:
            new_info = {
                "latest": new_commits[0].hexsha[:6],
                "commits": [
                    {
                        "hash": c.hexsha[:6],
                        "url": repo_url and os.path.join(repo_url, "commit", c.hexsha) or None,
                        "date": c.committed_datetime.isoformat(),
                        "message": c.message,
                    }
                    for c in new_commits
                ][:max_commits],
                "total": len(new_commits),
            }
    except Exception as err:
        logging.warning("Can't check for new version: %s", err)
        raise err

    return new_info


def get_version(folder):
    """return revision of a git folder"""
    try:
        repo = Repo(folder)  # app or lib dir
    except InvalidGitRepositoryError:
        logging.warning("Not a valid git repository for folder '%s', skipped for getting its version.", folder)
        return
    try:
        # Get URL from actual remote branch name that is being tracked.
        # do not assume that the active branch is tracking origin,
        # or if it is tracking anything, or if the alias origin exists
        remote_name = repo.active_branch.tracking_branch().remote_name
        url = repo.remote(remote_name).url
    except Exception:  # pylint: disable=W0702
        # it is possible that the active branch is not tracking anything
        url = None
    try:
        commit = repo.head.object.hexsha[:6]
        commitdate = repo.head.object.committed_datetime.isoformat()
    except Exception as err:
        logging.warning("can't determine app commit hash: %s", err)
        commit = "unknown"
        commitdate = "unknown"

    try:
        return {
            "branch": repo.active_branch.name,
            "commit": commit,
            "date": commitdate,
            "giturl": url,
        }
    except Exception as err:
        logging.warning("can't determine app version, assuming HEAD detached': %s", err)
        return {
            "branch": "HEAD detached",
            "commit": commit,
            "date": commitdate,
            "giturl": url,
        }


def set_versions(config, app_folder):
    """
    Propagate versions (git branch name) in config module.
    Also set app and biothings folder paths (though not
    exposed as a config param since they are lower-cased,
    see biothings.__init__.py, regex PARAM_PAT)
    """
    if not os.path.exists(app_folder):
        raise FileNotFoundError(f"'{app_folder}' application folder doesn't exist")
    # app_version: version of the API application
    if not hasattr(config, "APP_VERSION"):
        config.APP_VERSION = get_version(app_folder)
        config.app_folder = app_folder
    else:
        logging.info("app_version '%s' forced in configuration file", config.APP_VERSION)

    # biothings_version: version of BioThings SDK
    if not hasattr(config, "BIOTHINGS_VERSION"):
        import biothings

        # .../biothings.api/biothings/__init__.py
        bt_folder, _bt = os.path.split(os.path.split(os.path.realpath(biothings.__file__))[0])
        if not os.path.exists(bt_folder):
            raise FileNotFoundError(f"'{bt_folder}' biothings folder doesn't exist")
        assert _bt == "biothings", "Expectig 'biothings' dir in biothings lib path"
        config.BIOTHINGS_VERSION = get_version(bt_folder)
        config.biothings_folder = bt_folder
    else:
        logging.info("biothings_version '%s' forced in configuration file", config.BIOTHINGS_VERSION)

    logging.info(
        "Running app_version=%s with biothings_version=%s", repr(config.APP_VERSION), repr(config.BIOTHINGS_VERSION)
    )


def get_source_code_info(src_file):
    """
    Given a path to a source code, try to find information
    about repository, revision, URL pointing to that file, etc...
    Return None if nothing can be determined.
    Tricky cases:
      - src_file could refer to another repo, within current repo
        (namely a remote data plugin, cloned within the api's plugins folder
      - src_file could point to a folder, when for instance a dataplugin is
        analized. This is because we can't point to an uploader file since
        it's dynamically generated
    """
    # need to be absolute to build proper github URL
    abs_src_file = os.path.abspath(src_file)
    try:
        repo = Repo(abs_src_file, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        logging.exception("Can't find a github repository for file '%s'", src_file)
        return None
    try:
        gcmd = repo.git
        _hash = gcmd.rev_list(-1, repo.active_branch, abs_src_file)
        rel_src_file = abs_src_file.replace(repo.working_dir, "").strip("/")
        if not _hash:
            # seems to be a repo cloned within a repo, change directory
            curdir = os.path.abspath(os.curdir)
            try:
                if os.path.isdir(abs_src_file):
                    os.chdir(abs_src_file)
                    _hash = gcmd.rev_list(-1, repo.active_branch)
                else:
                    dirname, filename = os.path.split(abs_src_file)
                    os.chdir(dirname)
                    _hash = gcmd.rev_list(-1, repo.active_branch, filename)
                rel_src_file = ""  # will point to folder by commit hash
            finally:
                os.chdir(curdir)
        if _hash:
            short_hash = gcmd.rev_parse(_hash, short=7)
        else:
            logging.warning("Couldn't determine commit hash for file '%s'", src_file)
            _hash = None
            short_hash = None
        # could have more than one URLs for origin, only take first
        repo_url = next(repo.remote().urls)
        info = {
            "repo": repo_url,
            "commit": short_hash,
            "branch": repo.active_branch.name,
        }
        if os.path.isdir(src_file):
            info["folder"] = rel_src_file
        else:
            info["file"] = rel_src_file
        info = dict_sweep(info)
        # rebuild URL to that file
        if "github.com" in repo_url:
            info["url"] = os.path.join(re.sub(r"\.git$", "", repo_url), "tree", _hash, rel_src_file)

        return info

    except GitCommandError:
        logging.exception("Error while getting git information for file '%s'", src_file)
        return None
    except TypeError as err:
        # happens with biothings symlink, just ignore
        logging.debug("Can't determine source code info (but that's fine): %s", err)
        return None



"""
    Generate a customized BioThings API given a supported database.

    biothings.web.launcher: Launch web applications in different environments.
    biothings.web.applications: HTTP web application over data services below.
    biothings.web.services & query: Data services built on top of connections.
    biothings.web.connections: Elasticsearch, MongoDB and SQL database access.

"""



import orjson


def make_object(attr, value):
    """
    Create dictionary following the input dot notation and the value.

    Example::

        make_object('a.b.c', 100) --> {a:{b:{c:100}}}
        make_object(['a','b','c'], 100) --> {a:{b:{c:100}}}
    """
    attr_list = attr.split(".")

    s = ""
    for k in attr_list:
        s += '{"' + k + '":'

    # Old implementation using json module
    # s += json.dumps(value)
    # s += "}" * (len(attr_list))
    # return json.loads(s)

    # New implementation using orjson module
    s += orjson.dumps(value).decode("utf-8")  # decoding is necessary because orjson dumps into bytes
    s += "}" * (len(attr_list))
    return orjson.loads(s)


def merge_object(obj1, obj2):
    for k in obj2:
        try:
            if isinstance(obj2[k], dict):
                obj1[k] = merge_object(obj1[k], obj2[k])
            else:
                obj1[k] = obj2[k]
        except Exception:
            obj1[k] = obj2[k]
    return obj1


def parse_dot_fields(doc):
    """
    Example: parse_dot_fields({'a': 1, 'b.c': 2, 'b.a.c': 3}) should return {'a': 1, 'b': {'a': {'c': 3}, 'c': 2}}
    """
    dot_fields = []
    expanded_doc = {}
    for key in doc:
        if key.find(".") != -1:
            dot_fields.append(key)
            expanded_doc = merge_object(expanded_doc, make_object(key, doc[key]))
    doc.update(expanded_doc)
    for key in dot_fields:
        del doc[key]
    return doc


def compose_dot_fields_by_fields(doc, fields):
    """
    Reverse funtion of parse_dot_fields
    """
    res = None
    to_del = set()
    for k in fields:
        if k.find(".") != -1:
            if not res:
                import copy
                res = copy.deepcopy(doc)
            ks = k.split(".")
            broke = False
            if ks[0] in doc:
                t = doc[ks[0]]
                for e in ks[1:]:
                    if e in t:
                        t = t[e]
                    else:
                        broke = True
                        break
                if broke:
                    continue
                to_del.add(ks[0])
                res[k] = t

    for k in to_del:
        del res[k]

    return res if res else doc



from functools import partial

from biothings import config as btconfig
from biothings.utils.manager import BaseManager

logger = btconfig.logger


class AutoSnapshotCleanupManager(BaseManager):
    """This feature will add a new console command "auto_snapshot_cleanup", a new api "/auto_snapshot_cleanup".
    It is intended to allow automatically cleanup old snapshot, based on configuration.

    This feature can be configurated by using AUTO_SNAPSHOT_CLEANUP_CONFIG variable, to determine:
        - schedule: how frequency this task should run
        - days: how old a snapshot should be deleted

    AUTO_SNAPSHOT_CLEANUP_CONFIG = {
        "environment_name": {
            "schedule": "* 0 * * *",      # run daily at 0am UTC
            "keep": 3,                   # the number of most recent snapshots to keep in one group
            "group_by": "build_config",  # the attr of which its values form groups
            "extra_filters": {}          # a set of criterions to limit which snapshots are to be cleaned
        },
        ...
    }
    """

    DEFAULT_SCHEDULE = "* 0 * * *"  # run daily at 0am UTC

    def __init__(self, snapshot_manager, job_manager, *args, **kwargs):
        super().__init__(job_manager, *args, **kwargs)

        self.snapshot_manager = snapshot_manager

    def configure(self, conf=None):
        self.auto_snapshot_cleaner_config = conf or {}

        for env_name in self.snapshot_manager.register.keys():
            cleaner_config = self.auto_snapshot_cleaner_config.get(env_name)

            if not isinstance(cleaner_config, dict):
                logger.info(f"Snapshot environment: {env_name}: No cleaner config found!")
                continue

            schedule = cleaner_config.get("schedule") or self.DEFAULT_SCHEDULE
            keep = cleaner_config.get("keep")
            group_by = cleaner_config.get("group_by")
            extra_filters = cleaner_config.get("extra_filters")

            self.job_manager.submit(
                partial(
                    self.snapshot_manager.cleanup,
                    env=env_name,
                    keep=keep,
                    group_by=group_by,
                    dryrun=False,
                    **extra_filters,
                ),
                schedule=schedule,
            )
